# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect
import threading

import pytest

from swh.model.hashutil import hash_to_bytes

from swh.indexer.storage.interface import IndexerStorageInterface


def prepare_mimetypes_from(fossology_licenses):
    """Fossology license needs some consistent data in db to run.

    """
    mimetypes = []
    for c in fossology_licenses:
        mimetypes.append({
            'id': c['id'],
            'mimetype': 'text/plain',
            'encoding': 'utf-8',
            'indexer_configuration_id': c['indexer_configuration_id'],
        })
    return mimetypes


def endpoint(storage, endpoint_type, endpoint_name):
    return getattr(storage, endpoint_type + '_' + endpoint_name)


class StorageETypeTester:
    """Base class for testing a series of common behaviour between a bunch of
    endpoint types supported by an IndexerStorage.

    This is supposed to be inherited with the following class attributes:
    - endpoint_type
    - tool_name
    - example_data

    See below for example usage.
    """

    def test_missing(self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool_id = data.tools[self.tool_name]['id']

        # given 2 (hopefully) unknown objects
        query = [
            {
                'id': data.sha1_1,
                'indexer_configuration_id': tool_id,
            },
            {
                'id': data.sha1_2,
                'indexer_configuration_id': tool_id,
            }]

        # we expect these are both returned by the xxx_missing endpoint
        actual_missing = endpoint(storage, etype, 'missing')(query)
        assert list(actual_missing) == [
            data.sha1_1,
            data.sha1_2,
        ]

        # now, when we add one of them
        endpoint(storage, etype, 'add')([{
            'id': data.sha1_2,
            **self.example_data[0],
            'indexer_configuration_id': tool_id,
        }])

        # we expect only the other one returned
        actual_missing = endpoint(storage, etype, 'missing')(query)
        assert list(actual_missing) == [data.sha1_1]

    def test_add__drop_duplicate(self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool_id = data.tools[self.tool_name]['id']

        # add the first object
        data_v1 = {
            'id': data.sha1_2,
            **self.example_data[0],
            'indexer_configuration_id': tool_id,
        }
        endpoint(storage, etype, 'add')([data_v1])

        # should be able to retrieve it
        actual_data = list(endpoint(storage, etype, 'get')([data.sha1_2]))
        expected_data_v1 = [{
            'id': data.sha1_2,
            **self.example_data[0],
            'tool': data.tools[self.tool_name],
        }]
        assert actual_data == expected_data_v1

        # now if we add a modified version of the same object (same id)
        data_v2 = data_v1.copy()
        data_v2.update(self.example_data[1])
        endpoint(storage, etype, 'add')([data_v2])

        # we expect to retrieve the original data, not the modified one
        actual_data = list(endpoint(storage, etype, 'get')([data.sha1_2]))
        assert actual_data == expected_data_v1

    def test_add__update_in_place_duplicate(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        data_v1 = {
            'id': data.sha1_2,
            **self.example_data[0],
            'indexer_configuration_id': tool['id'],
        }

        # given
        endpoint(storage, etype, 'add')([data_v1])

        # when
        actual_data = list(endpoint(storage, etype, 'get')([data.sha1_2]))

        expected_data_v1 = [{
            'id': data.sha1_2,
            **self.example_data[0],
            'tool': tool,
        }]

        # then
        assert actual_data == expected_data_v1

        # given
        data_v2 = data_v1.copy()
        data_v2.update(self.example_data[1])

        endpoint(storage, etype, 'add')([data_v2], conflict_update=True)

        actual_data = list(endpoint(storage, etype, 'get')([data.sha1_2]))

        expected_data_v2 = [{
            'id': data.sha1_2,
            **self.example_data[1],
            'tool': tool,
        }]

        # data did change as the v2 was used to overwrite v1
        assert actual_data == expected_data_v2

    def test_add__update_in_place_deadlock(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        hashes = [
            hash_to_bytes(
                '34973274ccef6ab4dfaaf86599792fa9c3fe4{:03d}'.format(i))
            for i in range(1000)]

        data_v1 = [
            {
                'id': hash_,
                **self.example_data[0],
                'indexer_configuration_id': tool['id'],
            }
            for hash_ in hashes
        ]
        data_v2 = [
            {
                'id': hash_,
                **self.example_data[1],
                'indexer_configuration_id': tool['id'],
            }
            for hash_ in hashes
        ]

        # Remove one item from each, so that both queries have to succeed for
        # all items to be in the DB.
        data_v2a = data_v2[1:]
        data_v2b = list(reversed(data_v2[0:-1]))

        # given
        endpoint(storage, etype, 'add')(data_v1)

        # when
        actual_data = list(endpoint(storage, etype, 'get')(hashes))

        expected_data_v1 = [
            {
                'id': hash_,
                **self.example_data[0],
                'tool': tool,
            }
            for hash_ in hashes
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        def f1():
            endpoint(storage, etype, 'add')(data_v2a, conflict_update=True)

        def f2():
            endpoint(storage, etype, 'add')(data_v2b, conflict_update=True)

        t1 = threading.Thread(target=f1)
        t2 = threading.Thread(target=f2)
        t2.start()
        t1.start()

        t1.join()
        t2.join()

        actual_data = sorted(endpoint(storage, etype, 'get')(hashes),
                             key=lambda x: x['id'])

        expected_data_v2 = [
            {
                'id': hash_,
                **self.example_data[1],
                'tool': tool,
            }
            for hash_ in hashes
        ]

        assert actual_data == expected_data_v2

    def test_add__duplicate_twice(self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        data_rev1 = {
            'id': data.revision_id_2,
            **self.example_data[0],
            'indexer_configuration_id': tool['id']
        }

        data_rev2 = {
            'id': data.revision_id_2,
            **self.example_data[1],
            'indexer_configuration_id': tool['id']
        }

        # when
        endpoint(storage, etype, 'add')([data_rev1])

        with pytest.raises(ValueError):
            endpoint(storage, etype, 'add')(
                [data_rev2, data_rev2],
                conflict_update=True)

        # then
        actual_data = list(endpoint(storage, etype, 'get')(
            [data.revision_id_2, data.revision_id_1]))

        expected_data = [{
            'id': data.revision_id_2,
            **self.example_data[0],
            'tool': tool,
        }]
        assert actual_data == expected_data

    def test_get(self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        query = [data.sha1_2, data.sha1_1]
        data1 = {
            'id': data.sha1_2,
            **self.example_data[0],
            'indexer_configuration_id': tool['id'],
        }

        # when
        endpoint(storage, etype, 'add')([data1])

        # then
        actual_data = list(endpoint(storage, etype, 'get')(query))

        # then
        expected_data = [{
            'id': data.sha1_2,
            **self.example_data[0],
            'tool': tool,
        }]

        assert actual_data == expected_data


class TestIndexerStorageContentMimetypes(StorageETypeTester):
    """Test Indexer Storage content_mimetype related methods
    """
    endpoint_type = 'content_mimetype'
    tool_name = 'file'
    example_data = [
        {
            'mimetype': 'text/plain',
            'encoding': 'utf-8',
        },
        {
            'mimetype': 'text/html',
            'encoding': 'us-ascii',
        },
        ]

    def test_generate_content_mimetype_get_range_limit_none(
            self, swh_indexer_storage):
        """mimetype_get_range call with wrong limit input should fail"""
        storage = swh_indexer_storage
        with pytest.raises(ValueError) as e:
            storage.content_mimetype_get_range(
                start=None, end=None, indexer_configuration_id=None,
                limit=None)

        assert e.value.args == (
            'Development error: limit should not be None',)

    def test_generate_content_mimetype_get_range_no_limit(
            self, swh_indexer_storage_with_data):
        """mimetype_get_range returns mimetypes within range provided"""
        storage, data = swh_indexer_storage_with_data
        mimetypes = data.mimetypes

        # All ids from the db
        content_ids = sorted([c['id'] for c in mimetypes])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve mimetypes
        tool_id = mimetypes[0]['indexer_configuration_id']
        actual_result = storage.content_mimetype_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert len(mimetypes) == len(actual_ids)
        assert actual_next is None
        assert content_ids == actual_ids

    def test_generate_content_mimetype_get_range_limit(
            self, swh_indexer_storage_with_data):
        """mimetype_get_range paginates results if limit exceeded"""
        storage, data = swh_indexer_storage_with_data

        indexer_configuration_id = data.tools['file']['id']

        # input the list of sha1s we want from storage
        content_ids = sorted(
            [c['id'] for c in data.mimetypes])
        mimetypes = list(storage.content_mimetype_get(content_ids))
        assert len(mimetypes) == len(data.mimetypes)

        start = content_ids[0]
        end = content_ids[-1]
        # retrieve mimetypes limited to 10 results
        actual_result = storage.content_mimetype_get_range(
            start, end,
            indexer_configuration_id=indexer_configuration_id,
            limit=10)

        assert actual_result
        assert set(actual_result.keys()) == {'ids', 'next'}
        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert len(actual_ids) == 10
        assert actual_next is not None
        assert actual_next == content_ids[10]

        expected_mimetypes = content_ids[:10]
        assert expected_mimetypes == actual_ids

        # retrieve next part
        actual_result = storage.content_mimetype_get_range(
            start=end, end=end,
            indexer_configuration_id=indexer_configuration_id)
        assert set(actual_result.keys()) == {'ids', 'next'}
        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert actual_next is None
        expected_mimetypes = [content_ids[-1]]
        assert expected_mimetypes == actual_ids


class TestIndexerStorageContentLanguage(StorageETypeTester):
    """Test Indexer Storage content_language related methods
    """
    endpoint_type = 'content_language'
    tool_name = 'pygments'
    example_data = [
        {
            'lang': 'haskell',
        },
        {
            'lang': 'common-lisp',
        },
        ]


class TestIndexerStorageContentCTags(StorageETypeTester):
    """Test Indexer Storage content_ctags related methods
    """
    endpoint_type = 'content_ctags'
    tool_name = 'universal-ctags'
    example_data = [
        {
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 119,
                'lang': 'OCaml',
            }]
        },
        {
            'ctags': [
                {
                    'name': 'done',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'Python',
                },
                {
                    'name': 'main',
                    'kind': 'function',
                    'line': 119,
                    'lang': 'Python',
                }]
        },
        ]

    # the following tests are disabled because CTAGS behaves differently
    @pytest.mark.skip
    def test_add__drop_duplicate(self):
        pass

    @pytest.mark.skip
    def test_add__update_in_place_duplicate(self):
        pass

    @pytest.mark.skip
    def test_add__update_in_place_deadlock(self):
        pass

    @pytest.mark.skip
    def test_add__duplicate_twice(self):
        pass

    @pytest.mark.skip
    def test_get(self):
        pass

    def test_content_ctags_search(self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # 1. given
        tool = data.tools['universal-ctags']
        tool_id = tool['id']

        ctag1 = {
            'id': data.sha1_1,
            'indexer_configuration_id': tool_id,
            'ctags': [
                {
                    'name': 'hello',
                    'kind': 'function',
                    'line': 133,
                    'lang': 'Python',
                },
                {
                    'name': 'counter',
                    'kind': 'variable',
                    'line': 119,
                    'lang': 'Python',
                },
                {
                    'name': 'hello',
                    'kind': 'variable',
                    'line': 210,
                    'lang': 'Python',
                },
            ]
        }

        ctag2 = {
            'id': data.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [
                {
                    'name': 'hello',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'C',
                },
                {
                    'name': 'result',
                    'kind': 'variable',
                    'line': 120,
                    'lang': 'C',
                },
            ]
        }

        storage.content_ctags_add([ctag1, ctag2])

        # 1. when
        actual_ctags = list(storage.content_ctags_search('hello', limit=1))

        # 1. then
        assert actual_ctags == [
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'function',
                'line': 133,
                'lang': 'Python',
            }
        ]

        # 2. when
        actual_ctags = list(storage.content_ctags_search(
            'hello',
            limit=1,
            last_sha1=ctag1['id']))

        # 2. then
        assert actual_ctags == [
            {
                'id': ctag2['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 100,
                'lang': 'C',
            }
        ]

        # 3. when
        actual_ctags = list(storage.content_ctags_search('hello'))

        # 3. then
        assert actual_ctags == [
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'function',
                'line': 133,
                'lang': 'Python',
            },
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 210,
                'lang': 'Python',
            },
            {
                'id': ctag2['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 100,
                'lang': 'C',
            },
        ]

        # 4. when
        actual_ctags = list(storage.content_ctags_search('counter'))

        # then
        assert actual_ctags == [{
            'id': ctag1['id'],
            'tool': tool,
            'name': 'counter',
            'kind': 'variable',
            'line': 119,
            'lang': 'Python',
        }]

        # 5. when
        actual_ctags = list(storage.content_ctags_search('result', limit=1))

        # then
        assert actual_ctags == [{
            'id': ctag2['id'],
            'tool': tool,
            'name': 'result',
            'kind': 'variable',
            'line': 120,
            'lang': 'C',
        }]

    def test_content_ctags_search_no_result(self, swh_indexer_storage):
        storage = swh_indexer_storage
        actual_ctags = list(storage.content_ctags_search('counter'))

        assert not actual_ctags

    def test_content_ctags_add__add_new_ctags_added(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data

        # given
        tool = data.tools['universal-ctags']
        tool_id = tool['id']

        ctag_v1 = {
            'id': data.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
            }]
        }

        # given
        storage.content_ctags_add([ctag_v1])
        storage.content_ctags_add([ctag_v1])  # conflict does nothing

        # when
        actual_ctags = list(storage.content_ctags_get([data.sha1_2]))

        # then
        expected_ctags = [{
            'id': data.sha1_2,
            'name': 'done',
            'kind': 'variable',
            'line': 100,
            'lang': 'Scheme',
            'tool': tool,
        }]

        assert actual_ctags == expected_ctags

        # given
        ctag_v2 = ctag_v1.copy()
        ctag_v2.update({
            'ctags': [
                {
                    'name': 'defn',
                    'kind': 'function',
                    'line': 120,
                    'lang': 'Scheme',
                }
            ]
        })

        storage.content_ctags_add([ctag_v2])

        expected_ctags = [
            {
                'id': data.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool,
            }, {
                'id': data.sha1_2,
                'name': 'defn',
                'kind': 'function',
                'line': 120,
                'lang': 'Scheme',
                'tool': tool,
            }
        ]

        actual_ctags = list(storage.content_ctags_get(
            [data.sha1_2]))

        assert actual_ctags == expected_ctags

    def test_content_ctags_add__update_in_place(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool = data.tools['universal-ctags']
        tool_id = tool['id']

        ctag_v1 = {
            'id': data.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
            }]
        }

        # given
        storage.content_ctags_add([ctag_v1])

        # when
        actual_ctags = list(storage.content_ctags_get(
            [data.sha1_2]))

        # then
        expected_ctags = [
            {
                'id': data.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool
            }
        ]
        assert actual_ctags == expected_ctags

        # given
        ctag_v2 = ctag_v1.copy()
        ctag_v2.update({
            'ctags': [
                {
                    'name': 'done',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'Scheme',
                },
                {
                    'name': 'defn',
                    'kind': 'function',
                    'line': 120,
                    'lang': 'Scheme',
                }
            ]
        })

        storage.content_ctags_add([ctag_v2], conflict_update=True)

        actual_ctags = list(storage.content_ctags_get(
            [data.sha1_2]))

        # ctag did change as the v2 was used to overwrite v1
        expected_ctags = [
            {
                'id': data.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool,
            },
            {
                'id': data.sha1_2,
                'name': 'defn',
                'kind': 'function',
                'line': 120,
                'lang': 'Scheme',
                'tool': tool,
            }
        ]
        assert actual_ctags == expected_ctags


class TestIndexerStorageContentMetadata(StorageETypeTester):
    """Test Indexer Storage content_metadata related methods
    """
    tool_name = 'swh-metadata-detector'
    endpoint_type = 'content_metadata'
    example_data = [
        {
            'metadata': {
                'other': {},
                'codeRepository': {
                    'type': 'git',
                    'url': 'https://github.com/moranegg/metadata_test'
                },
                'description': 'Simple package.json test for indexer',
                'name': 'test_metadata',
                'version': '0.0.1'
            },
        },
        {
            'metadata': {
                'other': {},
                'name': 'test_metadata',
                'version': '0.0.1'
            },
        },
        ]


class TestIndexerStorageRevisionIntrinsicMetadata(StorageETypeTester):
    """Test Indexer Storage revision_intrinsic_metadata related methods
    """
    tool_name = 'swh-metadata-detector'
    endpoint_type = 'revision_intrinsic_metadata'
    example_data = [
        {
            'metadata': {
                'other': {},
                'codeRepository': {
                    'type': 'git',
                    'url': 'https://github.com/moranegg/metadata_test'
                },
                'description': 'Simple package.json test for indexer',
                'name': 'test_metadata',
                'version': '0.0.1'
            },
            'mappings': ['mapping1'],
        },
        {
            'metadata': {
                'other': {},
                'name': 'test_metadata',
                'version': '0.0.1'
            },
            'mappings': ['mapping2'],
        },
        ]

    def test_revision_intrinsic_metadata_delete(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]

        query = [data.sha1_2, data.sha1_1]
        data1 = {
            'id': data.sha1_2,
            **self.example_data[0],
            'indexer_configuration_id': tool['id'],
        }

        # when
        endpoint(storage, etype, 'add')([data1])
        endpoint(storage, etype, 'delete')([
            {
                'id': data.sha1_2,
                'indexer_configuration_id': tool['id'],
            }
        ])

        # then
        actual_data = list(endpoint(storage, etype, 'get')(query))

        # then
        assert not actual_data

    def test_revision_intrinsic_metadata_delete_nonexisting(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        etype = self.endpoint_type
        tool = data.tools[self.tool_name]
        endpoint(storage, etype, 'delete')([
            {
                'id': data.sha1_2,
                'indexer_configuration_id': tool['id'],
            }
        ])


class TestIndexerStorageContentFossologyLicence:
    def test_content_fossology_license_add__new_license_added(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool = data.tools['nomos']
        tool_id = tool['id']

        license_v1 = {
            'id': data.sha1_1,
            'licenses': ['Apache-2.0'],
            'indexer_configuration_id': tool_id,
        }

        # given
        storage.content_fossology_license_add([license_v1])
        # conflict does nothing
        storage.content_fossology_license_add([license_v1])

        # when
        actual_licenses = list(storage.content_fossology_license_get(
            [data.sha1_1]))

        # then
        expected_license = {
            data.sha1_1: [{
                'licenses': ['Apache-2.0'],
                'tool': tool,
            }]
        }
        assert actual_licenses == [expected_license]

        # given
        license_v2 = license_v1.copy()
        license_v2.update({
            'licenses': ['BSD-2-Clause'],
        })

        storage.content_fossology_license_add([license_v2])

        actual_licenses = list(storage.content_fossology_license_get(
            [data.sha1_1]))

        expected_license = {
            data.sha1_1: [{
                'licenses': ['Apache-2.0', 'BSD-2-Clause'],
                'tool': tool
            }]
        }

        # license did not change as the v2 was dropped.
        assert actual_licenses == [expected_license]

    def test_generate_content_fossology_license_get_range_limit_none(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        """license_get_range call with wrong limit input should fail"""
        with pytest.raises(ValueError) as e:
            storage.content_fossology_license_get_range(
                start=None, end=None, indexer_configuration_id=None,
                limit=None)

        assert e.value.args == (
            'Development error: limit should not be None',)

    def test_generate_content_fossology_license_get_range_no_limit(
            self, swh_indexer_storage_with_data):
        """license_get_range returns licenses within range provided"""
        storage, data = swh_indexer_storage_with_data
        # craft some consistent mimetypes
        fossology_licenses = data.fossology_licenses
        mimetypes = prepare_mimetypes_from(fossology_licenses)

        storage.content_mimetype_add(mimetypes, conflict_update=True)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        content_ids = sorted([c['id'] for c in fossology_licenses])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = storage.content_fossology_license_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert len(fossology_licenses) == len(actual_ids)
        assert actual_next is None
        assert content_ids == actual_ids

    def test_generate_content_fossology_license_get_range_no_limit_with_filter(
            self, swh_indexer_storage_with_data):
        """This filters non textual, then returns results within range"""
        storage, data = swh_indexer_storage_with_data
        fossology_licenses = data.fossology_licenses
        mimetypes = data.mimetypes

        # craft some consistent mimetypes
        _mimetypes = prepare_mimetypes_from(fossology_licenses)
        # add binary mimetypes which will get filtered out in results
        for m in mimetypes:
            _mimetypes.append({
                'mimetype': 'binary',
                **m,
            })

        storage.content_mimetype_add(_mimetypes, conflict_update=True)
        # add fossology_licenses to storage
        storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        content_ids = sorted([c['id'] for c in fossology_licenses])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = storage.content_fossology_license_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert len(fossology_licenses) == len(actual_ids)
        assert actual_next is None
        assert content_ids == actual_ids

    def test_generate_fossology_license_get_range_limit(
            self, swh_indexer_storage_with_data):
        """fossology_license_get_range paginates results if limit exceeded"""
        storage, data = swh_indexer_storage_with_data
        fossology_licenses = data.fossology_licenses

        # craft some consistent mimetypes
        mimetypes = prepare_mimetypes_from(fossology_licenses)

        # add fossology_licenses to storage
        storage.content_mimetype_add(mimetypes, conflict_update=True)
        storage.content_fossology_license_add(fossology_licenses)

        # input the list of sha1s we want from storage
        content_ids = sorted([c['id'] for c in fossology_licenses])
        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses limited to 3 results
        limited_results = len(fossology_licenses) - 1
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = storage.content_fossology_license_get_range(
            start, end,
            indexer_configuration_id=tool_id, limit=limited_results)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        assert limited_results == len(actual_ids)
        assert actual_next is not None
        assert actual_next == content_ids[-1]

        expected_fossology_licenses = content_ids[:-1]
        assert expected_fossology_licenses == actual_ids

        # retrieve next part
        actual_results2 = storage.content_fossology_license_get_range(
            start=end, end=end, indexer_configuration_id=tool_id)
        actual_ids2 = actual_results2['ids']
        actual_next2 = actual_results2['next']

        assert actual_next2 is None
        expected_fossology_licenses2 = [content_ids[-1]]
        assert expected_fossology_licenses2 == actual_ids2


class TestIndexerStorageOriginIntrinsicMetadata:
    def test_origin_intrinsic_metadata_get(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata = {
            'version': None,
            'name': None,
        }
        metadata_rev = {
            'id': data.revision_id_2,
            'metadata': metadata,
            'mappings': ['mapping1'],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin = {
            'id': data.origin_url_1,
            'metadata': metadata,
            'indexer_configuration_id': tool_id,
            'mappings': ['mapping1'],
            'from_revision': data.revision_id_2,
            }

        # when
        storage.revision_intrinsic_metadata_add([metadata_rev])
        storage.origin_intrinsic_metadata_add([metadata_origin])

        # then
        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1, 'no://where']))

        expected_metadata = [{
            'id': data.origin_url_1,
            'metadata': metadata,
            'tool': data.tools['swh-metadata-detector'],
            'from_revision': data.revision_id_2,
            'mappings': ['mapping1'],
        }]

        assert actual_metadata == expected_metadata

    def test_origin_intrinsic_metadata_delete(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata = {
            'version': None,
            'name': None,
        }
        metadata_rev = {
            'id': data.revision_id_2,
            'metadata': metadata,
            'mappings': ['mapping1'],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin = {
            'id': data.origin_url_1,
            'metadata': metadata,
            'indexer_configuration_id': tool_id,
            'mappings': ['mapping1'],
            'from_revision': data.revision_id_2,
            }
        metadata_origin2 = metadata_origin.copy()
        metadata_origin2['id'] = data.origin_url_2

        # when
        storage.revision_intrinsic_metadata_add([metadata_rev])
        storage.origin_intrinsic_metadata_add([
            metadata_origin, metadata_origin2])

        storage.origin_intrinsic_metadata_delete([
            {
                'id': data.origin_url_1,
                'indexer_configuration_id': tool_id
            }
        ])

        # then
        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1, data.origin_url_2, 'no://where']))
        for item in actual_metadata:
            item['indexer_configuration_id'] = item.pop('tool')['id']
        assert actual_metadata == [metadata_origin2]

    def test_origin_intrinsic_metadata_delete_nonexisting(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool_id = data.tools['swh-metadata-detector']['id']
        storage.origin_intrinsic_metadata_delete([
            {
                'id': data.origin_url_1,
                'indexer_configuration_id': tool_id
            }
        ])

    def test_origin_intrinsic_metadata_add_drop_duplicate(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata_v1 = {
            'version': None,
            'name': None,
        }
        metadata_rev_v1 = {
            'id': data.revision_id_1,
            'metadata': metadata_v1.copy(),
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin_v1 = {
            'id': data.origin_url_1,
            'metadata': metadata_v1.copy(),
            'indexer_configuration_id': tool_id,
            'mappings': [],
            'from_revision': data.revision_id_1,
        }

        # given
        storage.revision_intrinsic_metadata_add([metadata_rev_v1])
        storage.origin_intrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1, 'no://where']))

        expected_metadata_v1 = [{
            'id': data.origin_url_1,
            'metadata': metadata_v1,
            'tool': data.tools['swh-metadata-detector'],
            'from_revision': data.revision_id_1,
            'mappings': [],
        }]

        assert actual_metadata == expected_metadata_v1

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update({
            'name': 'test_metadata',
            'author': 'MG',
        })
        metadata_rev_v2 = metadata_rev_v1.copy()
        metadata_origin_v2 = metadata_origin_v1.copy()
        metadata_rev_v2['metadata'] = metadata_v2
        metadata_origin_v2['metadata'] = metadata_v2

        storage.revision_intrinsic_metadata_add([metadata_rev_v2])
        storage.origin_intrinsic_metadata_add([metadata_origin_v2])

        # then
        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1]))

        # metadata did not change as the v2 was dropped.
        assert actual_metadata == expected_metadata_v1

    def test_origin_intrinsic_metadata_add_update_in_place_duplicate(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata_v1 = {
            'version': None,
            'name': None,
        }
        metadata_rev_v1 = {
            'id': data.revision_id_2,
            'metadata': metadata_v1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin_v1 = {
            'id': data.origin_url_1,
            'metadata': metadata_v1.copy(),
            'indexer_configuration_id': tool_id,
            'mappings': [],
            'from_revision': data.revision_id_2,
        }

        # given
        storage.revision_intrinsic_metadata_add([metadata_rev_v1])
        storage.origin_intrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1]))

        # then
        expected_metadata_v1 = [{
            'id': data.origin_url_1,
            'metadata': metadata_v1,
            'tool': data.tools['swh-metadata-detector'],
            'from_revision': data.revision_id_2,
            'mappings': [],
        }]
        assert actual_metadata == expected_metadata_v1

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update({
            'name': 'test_update_duplicated_metadata',
            'author': 'MG',
        })
        metadata_rev_v2 = metadata_rev_v1.copy()
        metadata_origin_v2 = metadata_origin_v1.copy()
        metadata_rev_v2['metadata'] = metadata_v2
        metadata_origin_v2 = {
            'id': data.origin_url_1,
            'metadata': metadata_v2.copy(),
            'indexer_configuration_id': tool_id,
            'mappings': ['npm'],
            'from_revision': data.revision_id_1,
        }

        storage.revision_intrinsic_metadata_add(
                [metadata_rev_v2], conflict_update=True)
        storage.origin_intrinsic_metadata_add(
                [metadata_origin_v2], conflict_update=True)

        actual_metadata = list(storage.origin_intrinsic_metadata_get(
            [data.origin_url_1]))

        expected_metadata_v2 = [{
            'id': data.origin_url_1,
            'metadata': metadata_v2,
            'tool': data.tools['swh-metadata-detector'],
            'from_revision': data.revision_id_1,
            'mappings': ['npm'],
        }]

        # metadata did change as the v2 was used to overwrite v1
        assert actual_metadata == expected_metadata_v2

    def test_origin_intrinsic_metadata_add__update_in_place_deadlock(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        ids = list(range(10))

        example_data1 = {
            'metadata': {
                'version': None,
                'name': None,
            },
            'mappings': [],
        }
        example_data2 = {
            'metadata': {
                'version': 'v1.1.1',
                'name': 'foo',
            },
            'mappings': [],
        }

        metadata_rev_v1 = {
            'id': data.revision_id_2,
            'metadata': {
                'version': None,
                'name': None,
            },
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }

        data_v1 = [
            {
                'id': 'file:///tmp/origin%d' % id_,
                'from_revision': data.revision_id_2,
                **example_data1,
                'indexer_configuration_id': tool_id,
            }
            for id_ in ids
        ]
        data_v2 = [
            {
                'id': 'file:///tmp/origin%d' % id_,
                'from_revision': data.revision_id_2,
                **example_data2,
                'indexer_configuration_id': tool_id,
            }
            for id_ in ids
        ]

        # Remove one item from each, so that both queries have to succeed for
        # all items to be in the DB.
        data_v2a = data_v2[1:]
        data_v2b = list(reversed(data_v2[0:-1]))

        # given
        storage.revision_intrinsic_metadata_add([metadata_rev_v1])
        storage.origin_intrinsic_metadata_add(data_v1)

        # when
        origins = ['file:///tmp/origin%d' % i for i in ids]
        actual_data = list(storage.origin_intrinsic_metadata_get(origins))

        expected_data_v1 = [
            {
                'id': 'file:///tmp/origin%d' % id_,
                'from_revision': data.revision_id_2,
                **example_data1,
                'tool': data.tools['swh-metadata-detector'],
            }
            for id_ in ids
        ]

        # then
        assert actual_data == expected_data_v1

        # given
        def f1():
            storage.origin_intrinsic_metadata_add(
                data_v2a, conflict_update=True)

        def f2():
            storage.origin_intrinsic_metadata_add(
                data_v2b, conflict_update=True)

        t1 = threading.Thread(target=f1)
        t2 = threading.Thread(target=f2)
        t2.start()
        t1.start()

        t1.join()
        t2.join()

        actual_data = list(storage.origin_intrinsic_metadata_get(origins))

        expected_data_v2 = [
            {
                'id': 'file:///tmp/origin%d' % id_,
                'from_revision': data.revision_id_2,
                **example_data2,
                'tool': data.tools['swh-metadata-detector'],
            }
            for id_ in ids
        ]

        assert len(actual_data) == len(expected_data_v2)
        assert sorted(actual_data, key=lambda x: x['id']) == expected_data_v2

    def test_origin_intrinsic_metadata_add__duplicate_twice(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata = {
            'developmentStatus': None,
            'name': None,
        }
        metadata_rev = {
            'id': data.revision_id_2,
            'metadata': metadata,
            'mappings': ['mapping1'],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin = {
            'id': data.origin_url_1,
            'metadata': metadata,
            'indexer_configuration_id': tool_id,
            'mappings': ['mapping1'],
            'from_revision': data.revision_id_2,
            }

        # when
        storage.revision_intrinsic_metadata_add([metadata_rev])

        with pytest.raises(ValueError):
            storage.origin_intrinsic_metadata_add([
                metadata_origin, metadata_origin])

    def test_origin_intrinsic_metadata_search_fulltext(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        metadata1 = {
            'author': 'John Doe',
        }
        metadata1_rev = {
            'id': data.revision_id_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata1_origin = {
            'id': data.origin_url_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': data.revision_id_1,
        }
        metadata2 = {
            'author': 'Jane Doe',
        }
        metadata2_rev = {
            'id': data.revision_id_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata2_origin = {
            'id': data.origin_url_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': data.revision_id_2,
        }

        # when
        storage.revision_intrinsic_metadata_add([metadata1_rev])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.revision_intrinsic_metadata_add([metadata2_rev])
        storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = storage.origin_intrinsic_metadata_search_fulltext
        assert set([res['id'] for res in search(['Doe'])]) \
            == set([data.origin_url_1, data.origin_url_2])
        assert [res['id'] for res in search(['John', 'Doe'])] \
            == [data.origin_url_1]
        assert [res['id'] for res in search(['John'])] \
            == [data.origin_url_1]
        assert not list(search(['John', 'Jane']))

    def test_origin_intrinsic_metadata_search_fulltext_rank(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        # given
        tool_id = data.tools['swh-metadata-detector']['id']

        # The following authors have "Random Person" to add some more content
        # to the JSON data, to work around normalization quirks when there
        # are few words (rank/(1+ln(nb_words)) is very sensitive to nb_words
        # for small values of nb_words).
        metadata1 = {
            'author': [
                'Random Person',
                'John Doe',
                'Jane Doe',
            ]
        }
        metadata1_rev = {
            'id': data.revision_id_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata1_origin = {
            'id': data.origin_url_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': data.revision_id_1,
        }
        metadata2 = {
            'author': [
                'Random Person',
                'Jane Doe',
            ]
        }
        metadata2_rev = {
            'id': data.revision_id_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata2_origin = {
            'id': data.origin_url_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': data.revision_id_2,
        }

        # when
        storage.revision_intrinsic_metadata_add([metadata1_rev])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.revision_intrinsic_metadata_add([metadata2_rev])
        storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = storage.origin_intrinsic_metadata_search_fulltext
        assert [res['id'] for res in search(['Doe'])] \
            == [data.origin_url_1, data.origin_url_2]
        assert [res['id'] for res in search(['Doe'], limit=1)] \
            == [data.origin_url_1]
        assert [res['id'] for res in search(['John'])] \
            == [data.origin_url_1]
        assert [res['id'] for res in search(['Jane'])] \
            == [data.origin_url_2, data.origin_url_1]
        assert [res['id'] for res in search(['John', 'Jane'])] \
            == [data.origin_url_1]

    def _fill_origin_intrinsic_metadata(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool1_id = data.tools['swh-metadata-detector']['id']
        tool2_id = data.tools['swh-metadata-detector2']['id']

        metadata1 = {
            '@context': 'foo',
            'author': 'John Doe',
        }
        metadata1_rev = {
            'id': data.revision_id_1,
            'metadata': metadata1,
            'mappings': ['npm'],
            'indexer_configuration_id': tool1_id,
        }
        metadata1_origin = {
            'id': data.origin_url_1,
            'metadata': metadata1,
            'mappings': ['npm'],
            'indexer_configuration_id': tool1_id,
            'from_revision': data.revision_id_1,
        }
        metadata2 = {
            '@context': 'foo',
            'author': 'Jane Doe',
        }
        metadata2_rev = {
            'id': data.revision_id_2,
            'metadata': metadata2,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
        }
        metadata2_origin = {
            'id': data.origin_url_2,
            'metadata': metadata2,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
            'from_revision': data.revision_id_2,
        }
        metadata3 = {
            '@context': 'foo',
        }
        metadata3_rev = {
            'id': data.revision_id_3,
            'metadata': metadata3,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
        }
        metadata3_origin = {
            'id': data.origin_url_3,
            'metadata': metadata3,
            'mappings': ['pkg-info'],
            'indexer_configuration_id': tool2_id,
            'from_revision': data.revision_id_3,
        }

        storage.revision_intrinsic_metadata_add([metadata1_rev])
        storage.origin_intrinsic_metadata_add([metadata1_origin])
        storage.revision_intrinsic_metadata_add([metadata2_rev])
        storage.origin_intrinsic_metadata_add([metadata2_origin])
        storage.revision_intrinsic_metadata_add([metadata3_rev])
        storage.origin_intrinsic_metadata_add([metadata3_origin])

    def test_origin_intrinsic_metadata_search_by_producer(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        self._fill_origin_intrinsic_metadata(
            swh_indexer_storage_with_data)
        tool1 = data.tools['swh-metadata-detector']
        tool2 = data.tools['swh-metadata-detector2']
        endpoint = storage.origin_intrinsic_metadata_search_by_producer

        # test pagination
        # no 'page_token' param, return all origins
        result = endpoint(ids_only=True)
        assert result['origins'] \
            == [data.origin_url_1, data.origin_url_2, data.origin_url_3]
        assert 'next_page_token' not in result

        # 'page_token' is < than origin_1, return everything
        result = endpoint(page_token=data.origin_url_1[:-1], ids_only=True)
        assert result['origins'] \
            == [data.origin_url_1, data.origin_url_2, data.origin_url_3]
        assert 'next_page_token' not in result

        # 'page_token' is origin_3, return nothing
        result = endpoint(page_token=data.origin_url_3, ids_only=True)
        assert not result['origins']
        assert 'next_page_token' not in result

        # test limit argument
        result = endpoint(page_token=data.origin_url_1[:-1],
                          limit=2, ids_only=True)
        assert result['origins'] == [data.origin_url_1, data.origin_url_2]
        assert result['next_page_token'] == result['origins'][-1]

        result = endpoint(page_token=data.origin_url_1, limit=2, ids_only=True)
        assert result['origins'] == [data.origin_url_2, data.origin_url_3]
        assert 'next_page_token' not in result

        result = endpoint(page_token=data.origin_url_2, limit=2, ids_only=True)
        assert result['origins'] == [data.origin_url_3]
        assert 'next_page_token' not in result

        # test mappings filtering
        result = endpoint(mappings=['npm'], ids_only=True)
        assert result['origins'] == [data.origin_url_1, data.origin_url_2]
        assert 'next_page_token' not in result

        result = endpoint(mappings=['npm', 'gemspec'], ids_only=True)
        assert result['origins'] == [data.origin_url_1, data.origin_url_2]
        assert 'next_page_token' not in result

        result = endpoint(mappings=['gemspec'], ids_only=True)
        assert result['origins'] == [data.origin_url_2]
        assert 'next_page_token' not in result

        result = endpoint(mappings=['pkg-info'], ids_only=True)
        assert result['origins'] == [data.origin_url_3]
        assert 'next_page_token' not in result

        result = endpoint(mappings=['foobar'], ids_only=True)
        assert not result['origins']
        assert 'next_page_token' not in result

        # test pagination + mappings
        result = endpoint(mappings=['npm'], limit=1, ids_only=True)
        assert result['origins'] == [data.origin_url_1]
        assert result['next_page_token'] == result['origins'][-1]

        # test tool filtering
        result = endpoint(tool_ids=[tool1['id']], ids_only=True)
        assert result['origins'] == [data.origin_url_1]
        assert 'next_page_token' not in result

        result = endpoint(tool_ids=[tool2['id']], ids_only=True)
        assert sorted(result['origins']) \
            == [data.origin_url_2, data.origin_url_3]
        assert 'next_page_token' not in result

        result = endpoint(tool_ids=[tool1['id'], tool2['id']], ids_only=True)
        assert sorted(result['origins']) \
            == [data.origin_url_1, data.origin_url_2, data.origin_url_3]
        assert 'next_page_token' not in result

        # test ids_only=False
        assert endpoint(mappings=['gemspec'])['origins'] \
            == [{
                'id': data.origin_url_2,
                'metadata': {
                    '@context': 'foo',
                    'author': 'Jane Doe',
                },
                'mappings': ['npm', 'gemspec'],
                'tool': tool2,
                'from_revision': data.revision_id_2,
            }]

    def test_origin_intrinsic_metadata_stats(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        self._fill_origin_intrinsic_metadata(
            swh_indexer_storage_with_data)

        result = storage.origin_intrinsic_metadata_stats()
        assert result == {
            'per_mapping': {
                'gemspec': 1,
                'npm': 2,
                'pkg-info': 1,
                'codemeta': 0,
                'maven': 0,
            },
            'total': 3,
            'non_empty': 2,
        }


class TestIndexerStorageIndexerCondifuration:
    def test_indexer_configuration_add(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'some-unknown-tool',
            'tool_version': 'some-version',
            'tool_configuration': {"debian-package": "some-package"},
        }

        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool is None  # does not exist

        # add it
        actual_tools = list(storage.indexer_configuration_add([tool]))

        assert len(actual_tools) == 1
        actual_tool = actual_tools[0]
        assert actual_tool is not None  # now it exists
        new_id = actual_tool.pop('id')
        assert actual_tool == tool

        actual_tools2 = list(storage.indexer_configuration_add([tool]))
        actual_tool2 = actual_tools2[0]
        assert actual_tool2 is not None  # now it exists
        new_id2 = actual_tool2.pop('id')

        assert new_id == new_id2
        assert actual_tool == actual_tool2

    def test_indexer_configuration_add_multiple(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'some-unknown-tool',
            'tool_version': 'some-version',
            'tool_configuration': {"debian-package": "some-package"},
        }

        actual_tools = list(storage.indexer_configuration_add([tool]))
        assert len(actual_tools) == 1

        new_tools = [tool, {
            'tool_name': 'yet-another-tool',
            'tool_version': 'version',
            'tool_configuration': {},
        }]

        actual_tools = list(storage.indexer_configuration_add(new_tools))
        assert len(actual_tools) == 2

        # order not guaranteed, so we iterate over results to check
        for tool in actual_tools:
            _id = tool.pop('id')
            assert _id is not None
            assert tool in new_tools

    def test_indexer_configuration_get_missing(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'unknown-tool',
            'tool_version': '3.1.0rc2-31-ga2cbb8c',
            'tool_configuration': {"command_line": "nomossa <filepath>"},
        }

        actual_tool = storage.indexer_configuration_get(tool)

        assert actual_tool is None

    def test_indexer_configuration_get(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'nomos',
            'tool_version': '3.1.0rc2-31-ga2cbb8c',
            'tool_configuration': {"command_line": "nomossa <filepath>"},
        }

        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool

        expected_tool = tool.copy()
        del actual_tool['id']

        assert expected_tool == actual_tool

    def test_indexer_configuration_metadata_get_missing_context(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'swh-metadata-translator',
            'tool_version': '0.0.1',
            'tool_configuration': {"context": "unknown-context"},
        }

        actual_tool = storage.indexer_configuration_get(tool)

        assert actual_tool is None

    def test_indexer_configuration_metadata_get(
            self, swh_indexer_storage_with_data):
        storage, data = swh_indexer_storage_with_data
        tool = {
            'tool_name': 'swh-metadata-translator',
            'tool_version': '0.0.1',
            'tool_configuration': {"type": "local", "context": "NpmMapping"},
        }

        storage.indexer_configuration_add([tool])
        actual_tool = storage.indexer_configuration_get(tool)
        assert actual_tool

        expected_tool = tool.copy()
        expected_tool['id'] = actual_tool['id']

        assert expected_tool == actual_tool


class TestIndexerStorageMisc:
    """Misc endpoints tests for the IndexerStorage.
    """

    def test_check_config(self, swh_indexer_storage):
        storage = swh_indexer_storage
        assert storage.check_config(check_write=True)
        assert storage.check_config(check_write=False)

    def test_types(self, swh_indexer_storage):
        """Checks all methods of StorageInterface are implemented by this
        backend, and that they have the same signature."""
        # Create an instance of the protocol (which cannot be instantiated
        # directly, so this creates a subclass, then instantiates it)
        interface = type('_', (IndexerStorageInterface,), {})()

        assert 'content_mimetype_add' in dir(interface)

        missing_methods = []

        for meth_name in dir(interface):
            if meth_name.startswith('_'):
                continue
            interface_meth = getattr(interface, meth_name)
            try:
                concrete_meth = getattr(swh_indexer_storage, meth_name)
            except AttributeError:
                missing_methods.append(meth_name)
                continue

            expected_signature = inspect.signature(interface_meth)
            actual_signature = inspect.signature(concrete_meth)

            assert expected_signature == actual_signature, meth_name

        assert missing_methods == []



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import datetime
import functools
import random
from typing import Dict, Any
import unittest

from hypothesis import strategies

from swh.model import hashutil
from swh.model.hashutil import hash_to_bytes, hash_to_hex

from swh.indexer.storage import INDEXER_CFG_KEY

BASE_TEST_CONFIG: Dict[str, Dict[str, Any]] = {
    'storage': {
        'cls': 'memory',
    },
    'objstorage': {
        'cls': 'memory',
        'args': {
        },
    },
    INDEXER_CFG_KEY: {
        'cls': 'memory',
        'args': {
        },
    },
}

ORIGIN_VISITS = [
        {
            'type': 'git',
            'url': 'https://github.com/SoftwareHeritage/swh-storage'},
        {
            'type': 'ftp',
            'url': 'rsync://ftp.gnu.org/gnu/3dldf'},
        {
            'type': 'deposit',
            'url': 'https://forge.softwareheritage.org/source/jesuisgpl/'},
        {
            'type': 'pypi',
            'url': 'https://pypi.org/project/limnoria/'},
        {
            'type': 'svn',
            'url': 'http://0-512-md.googlecode.com/svn/'},
        {
            'type': 'git',
            'url': 'https://github.com/librariesio/yarn-parser'},
        {
            'type': 'git',
            'url': 'https://github.com/librariesio/yarn-parser.git'},
        ]

SNAPSHOTS = [
    {
        'origin': 'https://github.com/SoftwareHeritage/swh-storage',
        'branches': {
            b'refs/heads/add-revision-origin-cache': {
                'target': b'L[\xce\x1c\x88\x8eF\t\xf1"\x19\x1e\xfb\xc0'
                          b's\xe7/\xe9l\x1e',
                'target_type': 'revision'},
            b'refs/head/master': {
                'target': b'8K\x12\x00d\x03\xcc\xe4]bS\xe3\x8f{\xd7}'
                          b'\xac\xefrm',
                'target_type': 'revision'},
            b'HEAD': {
                'target': b'refs/head/master',
                'target_type': 'alias'},
            b'refs/tags/v0.0.103': {
                'target': b'\xb6"Im{\xfdLb\xb0\x94N\xea\x96m\x13x\x88+'
                          b'\x0f\xdd',
                'target_type': 'release'},
            }},
    {
        'origin': 'rsync://ftp.gnu.org/gnu/3dldf',
        'branches': {
            b'3DLDF-1.1.4.tar.gz': {
                'target': b'dJ\xfb\x1c\x91\xf4\x82B%]6\xa2\x90|\xd3\xfc'
                          b'"G\x99\x11',
                'target_type': 'revision'},
            b'3DLDF-2.0.2.tar.gz': {
                'target': b'\xb6\x0e\xe7\x9e9\xac\xaa\x19\x9e='
                          b'\xd1\xc5\x00\\\xc6\xfc\xe0\xa6\xb4V',
                'target_type': 'revision'},
            b'3DLDF-2.0.3-examples.tar.gz': {
                'target': b'!H\x19\xc0\xee\x82-\x12F1\xbd\x97'
                          b'\xfe\xadZ\x80\x80\xc1\x83\xff',
                'target_type': 'revision'},
            b'3DLDF-2.0.3.tar.gz': {
                'target': b'\x8e\xa9\x8e/\xea}\x9feF\xf4\x9f\xfd\xee'
                          b'\xcc\x1a\xb4`\x8c\x8by',
                'target_type': 'revision'},
            b'3DLDF-2.0.tar.gz': {
                'target': b'F6*\xff(?\x19a\xef\xb6\xc2\x1fv$S\xe3G'
                          b'\xd3\xd1m',
                'target_type': 'revision'}
            }},
    {
        'origin': 'https://forge.softwareheritage.org/source/jesuisgpl/',
        'branches': {
            b'master': {
                'target': b'\xe7n\xa4\x9c\x9f\xfb\xb7\xf76\x11\x08{'
                          b'\xa6\xe9\x99\xb1\x9e]q\xeb',
                'target_type': 'revision'}
        },
        'id': b"h\xc0\xd2a\x04\xd4~'\x8d\xd6\xbe\x07\xeda\xfa\xfbV"
              b"\x1d\r "},
    {
        'origin': 'https://pypi.org/project/limnoria/',
        'branches': {
            b'HEAD': {
                'target': b'releases/2018.09.09',
                'target_type': 'alias'},
            b'releases/2018.09.01': {
                'target': b'<\xee1(\xe8\x8d_\xc1\xc9\xa6rT\xf1\x1d'
                          b'\xbb\xdfF\xfdw\xcf',
                'target_type': 'revision'},
            b'releases/2018.09.09': {
                'target': b'\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8k'
                          b'A\x10\x9d\xc5\xfa2\xf8t',
                'target_type': 'revision'}},
        'id': b'{\xda\x8e\x84\x7fX\xff\x92\x80^\x93V\x18\xa3\xfay'
              b'\x12\x9e\xd6\xb3'},
    {
        'origin': 'http://0-512-md.googlecode.com/svn/',
        'branches': {
            b'master': {
                'target': b'\xe4?r\xe1,\x88\xab\xec\xe7\x9a\x87\xb8'
                          b'\xc9\xad#.\x1bw=\x18',
                'target_type': 'revision'}},
        'id': b'\xa1\xa2\x8c\n\xb3\x87\xa8\xf9\xe0a\x8c\xb7'
              b'\x05\xea\xb8\x1f\xc4H\xf4s'},
    {
        'origin': 'https://github.com/librariesio/yarn-parser',
        'branches': {
            b'HEAD': {
                'target': hash_to_bytes(
                    '8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
                'target_type': 'revision'}}},
    {
        'origin': 'https://github.com/librariesio/yarn-parser.git',
        'branches': {
            b'HEAD': {
                'target': hash_to_bytes(
                    '8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
                'target_type': 'revision'}}},
]


REVISIONS = [{
    'id': hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
    'message': 'Improve search functionality',
    'author': {
        'name': b'Andrew Nesbitt',
        'fullname': b'Andrew Nesbitt <andrewnez@gmail.com>',
        'email': b'andrewnez@gmail.com'
    },
    'committer': {
        'name': b'Andrew Nesbitt',
        'fullname': b'Andrew Nesbitt <andrewnez@gmail.com>',
        'email': b'andrewnez@gmail.com'
    },
    'committer_date': {
        'negative_utc': None,
        'offset': 120,
        'timestamp': {
            'microseconds': 0,
            'seconds': 1380883849
        }
    },
    'type': 'git',
    'synthetic': False,
    'date': {
        'negative_utc': False,
        'timestamp': {
            'seconds': 1487596456,
            'microseconds': 0
        },
        'offset': 0
    },
    'directory': b'10'
}]

DIRECTORY_ID = b'10'

DIRECTORY_ENTRIES = [{
    'name': b'index.js',
    'type': 'file',
    'target': b'abc',
    'perms': 33188,
    },
    {
    'name': b'package.json',
    'type': 'file',
    'target': b'cde',
    'perms': 33188,
    },
    {
    'name': b'.github',
    'type': 'dir',
    'target': b'11',
    'perms': 16384,
    }
]

SHA1_TO_LICENSES = {
    '01c9379dfc33803963d07c1ccc748d3fe4c96bb5': ['GPL'],
    '02fb2c89e14f7fab46701478c83779c7beb7b069': ['Apache2.0'],
    '103bc087db1d26afc3a0283f38663d081e9b01e6': ['MIT'],
    '688a5ef812c53907562fe379d4b3851e69c7cb15': ['AGPL'],
    'da39a3ee5e6b4b0d3255bfef95601890afd80709': [],
}


SHA1_TO_CTAGS = {
    '01c9379dfc33803963d07c1ccc748d3fe4c96bb5': [{
        'name': 'foo',
        'kind': 'str',
        'line': 10,
        'lang': 'bar',
    }],
    'd4c647f0fc257591cc9ba1722484229780d1c607': [{
        'name': 'let',
        'kind': 'int',
        'line': 100,
        'lang': 'haskell',
    }],
    '688a5ef812c53907562fe379d4b3851e69c7cb15': [{
        'name': 'symbol',
        'kind': 'float',
        'line': 99,
        'lang': 'python',
    }],
}


OBJ_STORAGE_DATA = {
    '01c9379dfc33803963d07c1ccc748d3fe4c96bb5': b'this is some text',
    '688a5ef812c53907562fe379d4b3851e69c7cb15': b'another text',
    '8986af901dd2043044ce8f0d8fc039153641cf17': b'yet another text',
    '02fb2c89e14f7fab46701478c83779c7beb7b069': b"""
    import unittest
    import logging
    from swh.indexer.mimetype import MimetypeIndexer
    from swh.indexer.tests.test_utils import MockObjStorage

    class MockStorage():
        def content_mimetype_add(self, mimetypes):
            self.state = mimetypes
            self.conflict_update = conflict_update

        def indexer_configuration_add(self, tools):
            return [{
                'id': 10,
            }]
    """,
    '103bc087db1d26afc3a0283f38663d081e9b01e6': b"""
        #ifndef __AVL__
        #define __AVL__

        typedef struct _avl_tree avl_tree;

        typedef struct _data_t {
          int content;
        } data_t;
    """,
    '93666f74f1cf635c8c8ac118879da6ec5623c410': b"""
    (should 'pygments (recognize 'lisp 'easily))

    """,
    '26a9f72a7c87cc9205725cfd879f514ff4f3d8d5': b"""
    {
        "name": "test_metadata",
        "version": "0.0.1",
        "description": "Simple package.json test for indexer",
        "repository": {
          "type": "git",
          "url": "https://github.com/moranegg/metadata_test"
      }
    }
    """,
    'd4c647f0fc257591cc9ba1722484229780d1c607': b"""
    {
      "version": "5.0.3",
      "name": "npm",
      "description": "a package manager for JavaScript",
      "keywords": [
        "install",
        "modules",
        "package manager",
        "package.json"
      ],
      "preferGlobal": true,
      "config": {
        "publishtest": false
      },
      "homepage": "https://docs.npmjs.com/",
      "author": "Isaac Z. Schlueter <i@izs.me> (http://blog.izs.me)",
      "repository": {
        "type": "git",
        "url": "https://github.com/npm/npm"
      },
      "bugs": {
        "url": "https://github.com/npm/npm/issues"
      },
      "dependencies": {
        "JSONStream": "~1.3.1",
        "abbrev": "~1.1.0",
        "ansi-regex": "~2.1.1",
        "ansicolors": "~0.3.2",
        "ansistyles": "~0.1.3"
      },
      "devDependencies": {
        "tacks": "~1.2.6",
        "tap": "~10.3.2"
      },
      "license": "Artistic-2.0"
    }

    """,
    'a7ab314d8a11d2c93e3dcf528ca294e7b431c449': b"""
    """,
    'da39a3ee5e6b4b0d3255bfef95601890afd80709': b'',
    # 626364
    hash_to_hex(b'bcd'): b'unimportant content for bcd',
    # 636465
    hash_to_hex(b'cde'): b"""
    {
      "name": "yarn-parser",
      "version": "1.0.0",
      "description": "Tiny web service for parsing yarn.lock files",
      "main": "index.js",
      "scripts": {
        "start": "node index.js",
        "test": "mocha"
      },
      "engines": {
        "node": "9.8.0"
      },
      "repository": {
        "type": "git",
        "url": "git+https://github.com/librariesio/yarn-parser.git"
      },
      "keywords": [
        "yarn",
        "parse",
        "lock",
        "dependencies"
      ],
      "author": "Andrew Nesbitt",
      "license": "AGPL-3.0",
      "bugs": {
        "url": "https://github.com/librariesio/yarn-parser/issues"
      },
      "homepage": "https://github.com/librariesio/yarn-parser#readme",
      "dependencies": {
        "@yarnpkg/lockfile": "^1.0.0",
        "body-parser": "^1.15.2",
        "express": "^4.14.0"
      },
      "devDependencies": {
        "chai": "^4.1.2",
        "mocha": "^5.2.0",
        "request": "^2.87.0",
        "test": "^0.6.0"
      }
    }

"""
}

YARN_PARSER_METADATA = {
    '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
    'url':
        'https://github.com/librariesio/yarn-parser#readme',
    'codeRepository':
        'git+git+https://github.com/librariesio/yarn-parser.git',
    'author': [{
        'type': 'Person',
        'name': 'Andrew Nesbitt'
    }],
    'license': 'https://spdx.org/licenses/AGPL-3.0',
    'version': '1.0.0',
    'description':
        "Tiny web service for parsing yarn.lock files",
    'issueTracker':
        'https://github.com/librariesio/yarn-parser/issues',
    'name': 'yarn-parser',
    'keywords': ['yarn', 'parse', 'lock', 'dependencies'],
    'type': 'SoftwareSourceCode',
}


json_dict_keys = strategies.one_of(
    strategies.characters(),
    strategies.just('type'),
    strategies.just('url'),
    strategies.just('name'),
    strategies.just('email'),
    strategies.just('@id'),
    strategies.just('@context'),
    strategies.just('repository'),
    strategies.just('license'),
    strategies.just('repositories'),
    strategies.just('licenses'),
)
"""Hypothesis strategy that generates strings, with an emphasis on those
that are often used as dictionary keys in metadata files."""


generic_json_document = strategies.recursive(
    strategies.none() | strategies.booleans() | strategies.floats() |
    strategies.characters(),
    lambda children: (
        strategies.lists(children, 1) |
        strategies.dictionaries(json_dict_keys, children, min_size=1)
    )
)
"""Hypothesis strategy that generates possible values for values of JSON
metadata files."""


def json_document_strategy(keys=None):
    """Generates an hypothesis strategy that generates metadata files
    for a JSON-based format that uses the given keys."""
    if keys is None:
        keys = strategies.characters()
    else:
        keys = strategies.one_of(map(strategies.just, keys))

    return strategies.dictionaries(keys, generic_json_document, min_size=1)


def _tree_to_xml(root, xmlns, data):
    def encode(s):
        "Skips unpaired surrogates generated by json_document_strategy"
        return s.encode('utf8', 'replace')

    def to_xml(data, indent=b' '):
        if data is None:
            return b''
        elif isinstance(data, (bool, str, int, float)):
            return indent + encode(str(data))
        elif isinstance(data, list):
            return b'\n'.join(to_xml(v, indent=indent) for v in data)
        elif isinstance(data, dict):
            lines = []
            for (key, value) in data.items():
                lines.append(indent + encode('<{}>'.format(key)))
                lines.append(to_xml(value, indent=indent+b' '))
                lines.append(indent + encode('</{}>'.format(key)))
            return b'\n'.join(lines)
        else:
            raise TypeError(data)

    return b'\n'.join([
        '<{} xmlns="{}">'.format(root, xmlns).encode(),
        to_xml(data),
        '</{}>'.format(root).encode(),
    ])


class TreeToXmlTest(unittest.TestCase):
    def test_leaves(self):
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', None),
            b'<root xmlns="http://example.com">\n\n</root>'
        )
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', True),
            b'<root xmlns="http://example.com">\n True\n</root>'
        )
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', 'abc'),
            b'<root xmlns="http://example.com">\n abc\n</root>'
        )
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', 42),
            b'<root xmlns="http://example.com">\n 42\n</root>'
        )
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', 3.14),
            b'<root xmlns="http://example.com">\n 3.14\n</root>'
        )

    def test_dict(self):
        self.assertIn(
            _tree_to_xml('root', 'http://example.com', {
                'foo': 'bar',
                'baz': 'qux'
            }),
            [
                b'<root xmlns="http://example.com">\n'
                b' <foo>\n  bar\n </foo>\n'
                b' <baz>\n  qux\n </baz>\n'
                b'</root>',
                b'<root xmlns="http://example.com">\n'
                b' <baz>\n  qux\n </baz>\n'
                b' <foo>\n  bar\n </foo>\n'
                b'</root>'
            ]
        )

    def test_list(self):
        self.assertEqual(
            _tree_to_xml('root', 'http://example.com', [
                {'foo': 'bar'},
                {'foo': 'baz'},
            ]),
            b'<root xmlns="http://example.com">\n'
            b' <foo>\n  bar\n </foo>\n'
            b' <foo>\n  baz\n </foo>\n'
            b'</root>'
        )


def xml_document_strategy(keys, root, xmlns):
    """Generates an hypothesis strategy that generates metadata files
    for an XML format that uses the given keys."""

    return strategies.builds(
        functools.partial(_tree_to_xml, root, xmlns),
        json_document_strategy(keys))


def filter_dict(d, keys):
    'return a copy of the dict with keys deleted'
    if not isinstance(keys, (list, tuple)):
        keys = (keys, )
    return dict((k, v) for (k, v) in d.items() if k not in keys)


def fill_obj_storage(obj_storage):
    """Add some content in an object storage."""
    for (obj_id, content) in OBJ_STORAGE_DATA.items():
        obj_storage.add(content, obj_id=hash_to_bytes(obj_id))


def fill_storage(storage):
    visit_types = {}
    for visit in ORIGIN_VISITS:
        storage.origin_add_one({'url': visit['url']})
        visit_types[visit['url']] = visit['type']
    for snap in SNAPSHOTS:
        origin_url = snap['origin']
        visit = storage.origin_visit_add(
            origin=origin_url,
            date=datetime.datetime.now(),
            type=visit_types[origin_url])
        snap_id = snap.get('id') or \
            bytes([random.randint(0, 255) for _ in range(32)])
        storage.snapshot_add([{
            'id': snap_id,
            'branches': snap['branches']
        }])
        storage.origin_visit_update(
            origin_url, visit['visit'], status='full', snapshot=snap_id)
    storage.revision_add(REVISIONS)

    contents = []
    for (obj_id, content) in OBJ_STORAGE_DATA.items():
        content_hashes = hashutil.MultiHash.from_data(content).digest()
        contents.append({
            'data': content,
            'length': len(content),
            'status': 'visible',
            'sha1': hash_to_bytes(obj_id),
            'sha1_git': hash_to_bytes(obj_id),
            'sha256': content_hashes['sha256'],
            'blake2s256': content_hashes['blake2s256']
        })
    storage.content_add(contents)
    storage.directory_add([{
        'id': DIRECTORY_ID,
        'entries': DIRECTORY_ENTRIES,
    }])


class CommonContentIndexerTest(metaclass=abc.ABCMeta):
    legacy_get_format = False
    """True if and only if the tested indexer uses the legacy format.
    see: https://forge.softwareheritage.org/T1433

    """
    def get_indexer_results(self, ids):
        """Override this for indexers that don't have a mock storage."""
        return self.indexer.idx_storage.state

    def assert_legacy_results_ok(self, sha1s, expected_results=None):
        # XXX old format, remove this when all endpoints are
        #     updated to the new one
        #     see: https://forge.softwareheritage.org/T1433
        sha1s = [sha1 if isinstance(sha1, bytes) else hash_to_bytes(sha1)
                 for sha1 in sha1s]
        actual_results = list(self.get_indexer_results(sha1s))

        if expected_results is None:
            expected_results = self.expected_results

        self.assertEqual(len(expected_results), len(actual_results),
                         (expected_results, actual_results))
        for indexed_data in actual_results:
            _id = indexed_data['id']
            expected_data = expected_results[hashutil.hash_to_hex(_id)].copy()
            expected_data['id'] = _id
            self.assertEqual(indexed_data, expected_data)

    def assert_results_ok(self, sha1s, expected_results=None):
        if self.legacy_get_format:
            self.assert_legacy_results_ok(sha1s, expected_results)
            return

        sha1s = [sha1 if isinstance(sha1, bytes) else hash_to_bytes(sha1)
                 for sha1 in sha1s]
        actual_results = list(self.get_indexer_results(sha1s))

        if expected_results is None:
            expected_results = self.expected_results

        self.assertEqual(len(expected_results), len(actual_results),
                         (expected_results, actual_results))
        for indexed_data in actual_results:
            (_id, indexed_data) = list(indexed_data.items())[0]
            expected_data = expected_results[hashutil.hash_to_hex(_id)].copy()
            expected_data = [expected_data]
            self.assertEqual(indexed_data, expected_data)

    def test_index(self):
        """Known sha1 have their data indexed

        """
        sha1s = [self.id0, self.id1, self.id2]

        # when
        self.indexer.run(sha1s, policy_update='update-dups')

        self.assert_results_ok(sha1s)

        # 2nd pass
        self.indexer.run(sha1s, policy_update='ignore-dups')

        self.assert_results_ok(sha1s)

    def test_index_one_unknown_sha1(self):
        """Unknown sha1 are not indexed"""
        sha1s = [self.id1,
                 '799a5ef812c53907562fe379d4b3851e69c7cb15',  # unknown
                 '800a5ef812c53907562fe379d4b3851e69c7cb15']  # unknown

        # when
        self.indexer.run(sha1s, policy_update='update-dups')

        # then
        expected_results = {
            k: v for k, v in self.expected_results.items() if k in sha1s
        }

        self.assert_results_ok(sha1s, expected_results)


class CommonContentIndexerRangeTest:
    """Allows to factorize tests on range indexer.

    """
    def setUp(self):
        self.contents = sorted(OBJ_STORAGE_DATA)

    def assert_results_ok(self, start, end, actual_results,
                          expected_results=None):
        if expected_results is None:
            expected_results = self.expected_results

        actual_results = list(actual_results)
        for indexed_data in actual_results:
            _id = indexed_data['id']
            assert isinstance(_id, bytes)
            indexed_data = indexed_data.copy()
            indexed_data['id'] = hash_to_hex(indexed_data['id'])
            self.assertEqual(indexed_data, expected_results[hash_to_hex(_id)])
            self.assertTrue(start <= _id <= end)
            _tool_id = indexed_data['indexer_configuration_id']
            self.assertEqual(_tool_id, self.indexer.tool['id'])

    def test__index_contents(self):
        """Indexing contents without existing data results in indexed data

        """
        _start, _end = [self.contents[0], self.contents[2]]  # output hex ids
        start, end = map(hashutil.hash_to_bytes, (_start, _end))
        # given
        actual_results = list(self.indexer._index_contents(
            start, end, indexed={}))

        self.assert_results_ok(start, end, actual_results)

    def test__index_contents_with_indexed_data(self):
        """Indexing contents with existing data results in less indexed data

        """
        _start, _end = [self.contents[0], self.contents[2]]  # output hex ids
        start, end = map(hashutil.hash_to_bytes, (_start, _end))
        data_indexed = [self.id0, self.id2]

        # given
        actual_results = self.indexer._index_contents(
            start, end, indexed=set(map(hash_to_bytes, data_indexed)))

        # craft the expected results
        expected_results = self.expected_results.copy()
        for already_indexed_key in data_indexed:
            expected_results.pop(already_indexed_key)

        self.assert_results_ok(
            start, end, actual_results, expected_results)

    def test_generate_content_get(self):
        """Optimal indexing should result in indexed data

        """
        _start, _end = [self.contents[0], self.contents[2]]  # output hex ids
        start, end = map(hashutil.hash_to_bytes, (_start, _end))

        # given
        actual_results = self.indexer.run(start, end)

        # then
        self.assertTrue(actual_results)

    def test_generate_content_get_input_as_bytes(self):
        """Optimal indexing should result in indexed data

        Input are in bytes here.

        """
        _start, _end = [self.contents[0], self.contents[2]]  # output hex ids
        start, end = map(hashutil.hash_to_bytes, (_start, _end))

        # given
        actual_results = self.indexer.run(  # checks the bytes input this time
            start, end, skip_existing=False)
        # no already indexed data so same result as prior test

        # then
        self.assertTrue(actual_results)

    def test_generate_content_get_no_result(self):
        """No result indexed returns False"""
        _start, _end = ['0000000000000000000000000000000000000000',
                        '0000000000000000000000000000000000000001']
        start, end = map(hashutil.hash_to_bytes, (_start, _end))
        # given
        actual_results = self.indexer.run(
            start, end, incremental=False)

        # then
        self.assertFalse(actual_results)



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import Mock, patch

from swh.indexer.journal_client import process_journal_objects


class JournalClientTest(unittest.TestCase):
    def testOneOriginVisit(self):
        mock_scheduler = Mock()
        messages = {
            'origin_visit': [
                {
                    'status': 'full',
                    'origin': 'file:///dev/zero',
                },
            ]
        }
        process_journal_objects(
            messages, scheduler=mock_scheduler,
            task_names={'origin_metadata': 'task-name'})
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]['next_run']
        self.assertEqual(args, ([
            {
                'arguments': {
                    'kwargs': {'policy_update': 'update-dups'},
                    'args': (['file:///dev/zero'],)
                },
                'policy': 'oneshot',
                'type': 'task-name'
            },
        ],))

    def testOriginVisitLegacy(self):
        mock_scheduler = Mock()
        messages = {
            'origin_visit': [
                {
                    'status': 'full',
                    'origin': {
                        'url': 'file:///dev/zero',
                    }
                },
            ]
        }
        process_journal_objects(
            messages, scheduler=mock_scheduler,
            task_names={'origin_metadata': 'task-name'})
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]['next_run']
        self.assertEqual(args, ([
            {
                'arguments': {
                    'kwargs': {'policy_update': 'update-dups'},
                    'args': (['file:///dev/zero'],)
                },
                'policy': 'oneshot',
                'type': 'task-name'
            },
        ],))

    def testOneOriginVisitBatch(self):
        mock_scheduler = Mock()
        messages = {
            'origin_visit': [
                {
                    'status': 'full',
                    'origin': 'file:///dev/zero',
                },
                {
                    'status': 'full',
                    'origin': 'file:///tmp/foobar',
                },
            ]
        }
        process_journal_objects(
            messages, scheduler=mock_scheduler,
            task_names={'origin_metadata': 'task-name'})
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]['next_run']
        self.assertEqual(args, ([
            {
                'arguments': {
                    'kwargs': {'policy_update': 'update-dups'},
                    'args': (['file:///dev/zero', 'file:///tmp/foobar'],)
                },
                'policy': 'oneshot',
                'type': 'task-name'
            },
        ],))

    @patch('swh.indexer.journal_client.MAX_ORIGINS_PER_TASK', 2)
    def testOriginVisitBatches(self):
        mock_scheduler = Mock()
        messages = {
            'origin_visit': [
                {
                    'status': 'full',
                    'origin': 'file:///dev/zero',
                },
                {
                    'status': 'full',
                    'origin': 'file:///tmp/foobar',
                },
                {
                    'status': 'full',
                    'origin': 'file:///tmp/spamegg',
                },
            ]
        }
        process_journal_objects(
            messages, scheduler=mock_scheduler,
            task_names={'origin_metadata': 'task-name'})
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]['next_run']
        del args[0][1]['next_run']
        self.assertEqual(args, ([
            {
                'arguments': {
                    'kwargs': {'policy_update': 'update-dups'},
                    'args': (['file:///dev/zero', 'file:///tmp/foobar'],)
                },
                'policy': 'oneshot',
                'type': 'task-name'
            },
            {
                'arguments': {
                    'kwargs': {'policy_update': 'update-dups'},
                    'args': (['file:///tmp/spamegg'],)
                },
                'policy': 'oneshot',
                'type': 'task-name'
            },
        ],))



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.indexer.storage import get_indexer_storage

from .test_storage import *  # noqa


@pytest.fixture
def swh_indexer_storage():
    storage_config = {
        'cls': 'memory',
        'args': {
        },
    }
    return get_indexer_storage(**storage_config)



import collections

import click

from . import maven, npm, codemeta, python, ruby

MAPPINGS = {
    'CodemetaMapping': codemeta.CodemetaMapping,
    'MavenMapping': maven.MavenMapping,
    'NpmMapping': npm.NpmMapping,
    'PythonPkginfoMapping': python.PythonPkginfoMapping,
    'GemspecMapping': ruby.GemspecMapping,
}


def list_terms():
    """Returns a dictionary with all supported CodeMeta terms as keys,
    and the mappings that support each of them as values."""
    d = collections.defaultdict(set)
    for mapping in MAPPINGS.values():
        for term in mapping.supported_terms():
            d[term].add(mapping)
    return d


@click.command()
@click.argument('mapping_name')
@click.argument('file_name')
def main(mapping_name, file_name):
    from pprint import pprint
    with open(file_name, 'rb') as fd:
        file_content = fd.read()
    res = MAPPINGS[mapping_name]().translate(file_content)
    pprint(res)


if __name__ == '__main__':
    main()



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import json
import psycopg2

from collections import defaultdict

from swh.storage.common import db_transaction_generator, db_transaction
from swh.storage.exc import StorageDBError
from .db import Db

from . import converters


INDEXER_CFG_KEY = 'indexer_storage'


MAPPING_NAMES = ['codemeta', 'gemspec', 'maven', 'npm', 'pkg-info']


def get_indexer_storage(cls, args):
    """Get an indexer storage object of class `storage_class` with
    arguments `storage_args`.

    Args:
        cls (str): storage's class, either 'local' or 'remote'
        args (dict): dictionary of arguments passed to the
            storage class constructor

    Returns:
        an instance of swh.indexer's storage (either local or remote)

    Raises:
        ValueError if passed an unknown storage class.

    """
    if cls == 'remote':
        from .api.client import RemoteStorage as IndexerStorage
    elif cls == 'local':
        from . import IndexerStorage
    elif cls == 'memory':
        from .in_memory import IndexerStorage
    else:
        raise ValueError('Unknown indexer storage class `%s`' % cls)

    return IndexerStorage(**args)


def _check_id_duplicates(data):
    """
    If any two dictionaries in `data` have the same id, raises
    a `ValueError`.

    Values associated to the key must be hashable.

    Args:
        data (List[dict]): List of dictionaries to be inserted

    >>> _check_id_duplicates([
    ...     {'id': 'foo', 'data': 'spam'},
    ...     {'id': 'bar', 'data': 'egg'},
    ... ])
    >>> _check_id_duplicates([
    ...     {'id': 'foo', 'data': 'spam'},
    ...     {'id': 'foo', 'data': 'egg'},
    ... ])
    Traceback (most recent call last):
      ...
    ValueError: The same id is present more than once.
    """
    if len({item['id'] for item in data}) < len(data):
        raise ValueError('The same id is present more than once.')


class IndexerStorage:
    """SWH Indexer Storage

    """
    def __init__(self, db, min_pool_conns=1, max_pool_conns=10):
        """
        Args:
            db_conn: either a libpq connection string, or a psycopg2 connection

        """
        try:
            if isinstance(db, psycopg2.extensions.connection):
                self._pool = None
                self._db = Db(db)
            else:
                self._pool = psycopg2.pool.ThreadedConnectionPool(
                    min_pool_conns, max_pool_conns, db
                )
                self._db = None
        except psycopg2.OperationalError as e:
            raise StorageDBError(e)

    def get_db(self):
        if self._db:
            return self._db
        return Db.from_pool(self._pool)

    def put_db(self, db):
        if db is not self._db:
            db.put_conn()

    @db_transaction()
    def check_config(self, *, check_write, db=None, cur=None):
        # Check permissions on one of the tables
        if check_write:
            check = 'INSERT'
        else:
            check = 'SELECT'

        cur.execute(
            "select has_table_privilege(current_user, 'content_mimetype', %s)",  # noqa
            (check,)
        )
        return cur.fetchone()[0]

    @db_transaction_generator()
    def content_mimetype_missing(self, mimetypes, db=None, cur=None):
        for obj in db.content_mimetype_missing_from_list(mimetypes, cur):
            yield obj[0]

    def _content_get_range(self, content_type, start, end,
                           indexer_configuration_id, limit=1000,
                           with_textual_data=False,
                           db=None, cur=None):
        if limit is None:
            raise ValueError('Development error: limit should not be None')
        if content_type not in db.content_indexer_names:
            err = 'Development error: Wrong type. Should be one of [%s]' % (
                ','.join(db.content_indexer_names))
            raise ValueError(err)

        ids = []
        next_id = None
        for counter, obj in enumerate(db.content_get_range(
                content_type, start, end, indexer_configuration_id,
                limit=limit+1, with_textual_data=with_textual_data, cur=cur)):
            _id = obj[0]
            if counter >= limit:
                next_id = _id
                break

            ids.append(_id)

        return {
            'ids': ids,
            'next': next_id
        }

    @db_transaction()
    def content_mimetype_get_range(self, start, end, indexer_configuration_id,
                                   limit=1000, db=None, cur=None):
        return self._content_get_range('mimetype', start, end,
                                       indexer_configuration_id, limit=limit,
                                       db=db, cur=cur)

    @db_transaction()
    def content_mimetype_add(self, mimetypes, conflict_update=False, db=None,
                             cur=None):
        _check_id_duplicates(mimetypes)
        mimetypes.sort(key=lambda m: m['id'])
        db.mktemp_content_mimetype(cur)
        db.copy_to(mimetypes, 'tmp_content_mimetype',
                   ['id', 'mimetype', 'encoding', 'indexer_configuration_id'],
                   cur)
        db.content_mimetype_add_from_temp(conflict_update, cur)

    @db_transaction_generator()
    def content_mimetype_get(self, ids, db=None, cur=None):
        for c in db.content_mimetype_get_from_list(ids, cur):
            yield converters.db_to_mimetype(
                dict(zip(db.content_mimetype_cols, c)))

    @db_transaction_generator()
    def content_language_missing(self, languages, db=None, cur=None):
        for obj in db.content_language_missing_from_list(languages, cur):
            yield obj[0]

    @db_transaction_generator()
    def content_language_get(self, ids, db=None, cur=None):
        for c in db.content_language_get_from_list(ids, cur):
            yield converters.db_to_language(
                dict(zip(db.content_language_cols, c)))

    @db_transaction()
    def content_language_add(self, languages, conflict_update=False, db=None,
                             cur=None):
        _check_id_duplicates(languages)
        languages.sort(key=lambda m: m['id'])
        db.mktemp_content_language(cur)
        # empty language is mapped to 'unknown'
        db.copy_to(
            ({
                'id': l['id'],
                'lang': 'unknown' if not l['lang'] else l['lang'],
                'indexer_configuration_id': l['indexer_configuration_id'],
            } for l in languages),
            'tmp_content_language',
            ['id', 'lang', 'indexer_configuration_id'], cur)

        db.content_language_add_from_temp(conflict_update, cur)

    @db_transaction_generator()
    def content_ctags_missing(self, ctags, db=None, cur=None):
        for obj in db.content_ctags_missing_from_list(ctags, cur):
            yield obj[0]

    @db_transaction_generator()
    def content_ctags_get(self, ids, db=None, cur=None):
        for c in db.content_ctags_get_from_list(ids, cur):
            yield converters.db_to_ctags(dict(zip(db.content_ctags_cols, c)))

    @db_transaction()
    def content_ctags_add(self, ctags, conflict_update=False, db=None,
                          cur=None):
        _check_id_duplicates(ctags)
        ctags.sort(key=lambda m: m['id'])

        def _convert_ctags(__ctags):
            """Convert ctags dict to list of ctags.

            """
            for ctags in __ctags:
                yield from converters.ctags_to_db(ctags)

        db.mktemp_content_ctags(cur)
        db.copy_to(list(_convert_ctags(ctags)),
                   tblname='tmp_content_ctags',
                   columns=['id', 'name', 'kind', 'line',
                            'lang', 'indexer_configuration_id'],
                   cur=cur)

        db.content_ctags_add_from_temp(conflict_update, cur)

    @db_transaction_generator()
    def content_ctags_search(self, expression,
                             limit=10, last_sha1=None, db=None, cur=None):
        for obj in db.content_ctags_search(expression, last_sha1, limit,
                                           cur=cur):
            yield converters.db_to_ctags(dict(zip(db.content_ctags_cols, obj)))

    @db_transaction_generator()
    def content_fossology_license_get(self, ids, db=None, cur=None):
        d = defaultdict(list)
        for c in db.content_fossology_license_get_from_list(ids, cur):
            license = dict(zip(db.content_fossology_license_cols, c))

            id_ = license['id']
            d[id_].append(converters.db_to_fossology_license(license))

        for id_, facts in d.items():
            yield {id_: facts}

    @db_transaction()
    def content_fossology_license_add(self, licenses, conflict_update=False,
                                      db=None, cur=None):
        _check_id_duplicates(licenses)
        licenses.sort(key=lambda m: m['id'])
        db.mktemp_content_fossology_license(cur)
        db.copy_to(
            ({
                'id': sha1['id'],
                'indexer_configuration_id': sha1['indexer_configuration_id'],
                'license': license,
              } for sha1 in licenses
                for license in sha1['licenses']),
            tblname='tmp_content_fossology_license',
            columns=['id', 'license', 'indexer_configuration_id'],
            cur=cur)
        db.content_fossology_license_add_from_temp(conflict_update, cur)

    @db_transaction()
    def content_fossology_license_get_range(
            self, start, end, indexer_configuration_id,
            limit=1000, db=None, cur=None):
        return self._content_get_range('fossology_license', start, end,
                                       indexer_configuration_id, limit=limit,
                                       with_textual_data=True, db=db, cur=cur)

    @db_transaction_generator()
    def content_metadata_missing(self, metadata, db=None, cur=None):
        for obj in db.content_metadata_missing_from_list(metadata, cur):
            yield obj[0]

    @db_transaction_generator()
    def content_metadata_get(self, ids, db=None, cur=None):
        for c in db.content_metadata_get_from_list(ids, cur):
            yield converters.db_to_metadata(
                dict(zip(db.content_metadata_cols, c)))

    @db_transaction()
    def content_metadata_add(self, metadata, conflict_update=False, db=None,
                             cur=None):
        _check_id_duplicates(metadata)
        metadata.sort(key=lambda m: m['id'])

        db.mktemp_content_metadata(cur)

        db.copy_to(metadata, 'tmp_content_metadata',
                   ['id', 'metadata', 'indexer_configuration_id'],
                   cur)
        db.content_metadata_add_from_temp(conflict_update, cur)

    @db_transaction_generator()
    def revision_intrinsic_metadata_missing(self, metadata, db=None, cur=None):
        for obj in db.revision_intrinsic_metadata_missing_from_list(
                metadata, cur):
            yield obj[0]

    @db_transaction_generator()
    def revision_intrinsic_metadata_get(self, ids, db=None, cur=None):
        for c in db.revision_intrinsic_metadata_get_from_list(ids, cur):
            yield converters.db_to_metadata(
                dict(zip(db.revision_intrinsic_metadata_cols, c)))

    @db_transaction()
    def revision_intrinsic_metadata_add(self, metadata, conflict_update=False,
                                        db=None, cur=None):
        _check_id_duplicates(metadata)
        metadata.sort(key=lambda m: m['id'])

        db.mktemp_revision_intrinsic_metadata(cur)

        db.copy_to(metadata, 'tmp_revision_intrinsic_metadata',
                   ['id', 'metadata', 'mappings',
                    'indexer_configuration_id'],
                   cur)
        db.revision_intrinsic_metadata_add_from_temp(conflict_update, cur)

    @db_transaction()
    def revision_intrinsic_metadata_delete(self, entries, db=None, cur=None):
        db.revision_intrinsic_metadata_delete(entries, cur)

    @db_transaction_generator()
    def origin_intrinsic_metadata_get(self, ids, db=None, cur=None):
        for c in db.origin_intrinsic_metadata_get_from_list(ids, cur):
            yield converters.db_to_metadata(
                dict(zip(db.origin_intrinsic_metadata_cols, c)))

    @db_transaction()
    def origin_intrinsic_metadata_add(self, metadata,
                                      conflict_update=False, db=None,
                                      cur=None):
        _check_id_duplicates(metadata)
        metadata.sort(key=lambda m: m['id'])

        db.mktemp_origin_intrinsic_metadata(cur)

        db.copy_to(metadata, 'tmp_origin_intrinsic_metadata',
                   ['id', 'metadata',
                    'indexer_configuration_id',
                    'from_revision', 'mappings'],
                   cur)
        db.origin_intrinsic_metadata_add_from_temp(conflict_update, cur)

    @db_transaction()
    def origin_intrinsic_metadata_delete(
            self, entries, db=None, cur=None):
        db.origin_intrinsic_metadata_delete(entries, cur)

    @db_transaction_generator()
    def origin_intrinsic_metadata_search_fulltext(
            self, conjunction, limit=100, db=None, cur=None):
        for c in db.origin_intrinsic_metadata_search_fulltext(
                conjunction, limit=limit, cur=cur):
            yield converters.db_to_metadata(
                dict(zip(db.origin_intrinsic_metadata_cols, c)))

    @db_transaction()
    def origin_intrinsic_metadata_search_by_producer(
            self, page_token='', limit=100, ids_only=False,
            mappings=None, tool_ids=None,
            db=None, cur=None):
        assert isinstance(page_token, str)
        # we go to limit+1 to check whether we should add next_page_token in
        # the response
        res = db.origin_intrinsic_metadata_search_by_producer(
            page_token, limit + 1, ids_only, mappings, tool_ids, cur)
        result = {}
        if ids_only:
            result['origins'] = [origin for (origin,) in res]
            if len(result['origins']) > limit:
                result['origins'][limit:] = []
                result['next_page_token'] = result['origins'][-1]
        else:
            result['origins'] = [converters.db_to_metadata(
                dict(zip(db.origin_intrinsic_metadata_cols, c)))for c in res]
            if len(result['origins']) > limit:
                result['origins'][limit:] = []
                result['next_page_token'] = result['origins'][-1]['id']
        return result

    @db_transaction()
    def origin_intrinsic_metadata_stats(
            self, db=None, cur=None):
        mapping_names = [m for m in MAPPING_NAMES]
        select_parts = []

        # Count rows for each mapping
        for mapping_name in mapping_names:
            select_parts.append((
                "sum(case when (mappings @> ARRAY['%s']) "
                "         then 1 else 0 end)"
                ) % mapping_name)

        # Total
        select_parts.append("sum(1)")

        # Rows whose metadata has at least one key that is not '@context'
        select_parts.append(
            "sum(case when ('{}'::jsonb @> (metadata - '@context')) "
            "         then 0 else 1 end)")
        cur.execute('select ' + ', '.join(select_parts)
                    + ' from origin_intrinsic_metadata')
        results = dict(zip(mapping_names + ['total', 'non_empty'],
                           cur.fetchone()))
        return {
            'total': results.pop('total'),
            'non_empty': results.pop('non_empty'),
            'per_mapping': results,
        }

    @db_transaction_generator()
    def indexer_configuration_add(self, tools, db=None, cur=None):
        db.mktemp_indexer_configuration(cur)
        db.copy_to(tools, 'tmp_indexer_configuration',
                   ['tool_name', 'tool_version', 'tool_configuration'],
                   cur)

        tools = db.indexer_configuration_add_from_temp(cur)
        for line in tools:
            yield dict(zip(db.indexer_configuration_cols, line))

    @db_transaction()
    def indexer_configuration_get(self, tool, db=None, cur=None):
        tool_conf = tool['tool_configuration']
        if isinstance(tool_conf, dict):
            tool_conf = json.dumps(tool_conf)
        idx = db.indexer_configuration_get(tool['tool_name'],
                                           tool['tool_version'],
                                           tool_conf)
        if not idx:
            return None
        return dict(zip(db.indexer_configuration_cols, idx))



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil

from swh.core.db import BaseDb
from swh.core.db.db_utils import execute_values_generator, stored_procedure


class Db(BaseDb):
    """Proxy to the SWH Indexer DB, with wrappers around stored procedures

    """
    content_mimetype_hash_keys = ['id', 'indexer_configuration_id']

    def _missing_from_list(self, table, data, hash_keys, cur=None):
        """Read from table the data with hash_keys that are missing.

        Args:
            table (str): Table name (e.g content_mimetype, content_language,
              etc...)
            data (dict): Dict of data to read from
            hash_keys ([str]): List of keys to read in the data dict.

        Yields:
            The data which is missing from the db.

        """
        cur = self._cursor(cur)
        keys = ', '.join(hash_keys)
        equality = ' AND '.join(
            ('t.%s = c.%s' % (key, key)) for key in hash_keys
        )
        yield from execute_values_generator(
            cur, """
            select %s from (values %%s) as t(%s)
            where not exists (
                select 1 from %s c
                where %s
            )
            """ % (keys, keys, table, equality),
            (tuple(m[k] for k in hash_keys) for m in data)
        )

    def content_mimetype_missing_from_list(self, mimetypes, cur=None):
        """List missing mimetypes.

        """
        yield from self._missing_from_list(
            'content_mimetype', mimetypes, self.content_mimetype_hash_keys,
            cur=cur)

    content_mimetype_cols = [
        'id', 'mimetype', 'encoding',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    @stored_procedure('swh_mktemp_content_mimetype')
    def mktemp_content_mimetype(self, cur=None): pass

    def content_mimetype_add_from_temp(self, conflict_update, cur=None):
        self._cursor(cur).execute("SELECT swh_content_mimetype_add(%s)",
                                  (conflict_update, ))

    def _convert_key(self, key, main_table='c'):
        """Convert keys according to specific use in the module.

        Args:
            key (str): Key expression to change according to the alias
              used in the query
            main_table (str): Alias to use for the main table. Default
              to c for content_{something}.

        Expected:
            Tables content_{something} being aliased as 'c' (something
            in {language, mimetype, ...}), table indexer_configuration
            being aliased as 'i'.

        """
        if key == 'id':
            return '%s.id' % main_table
        elif key == 'tool_id':
            return 'i.id as tool_id'
        elif key == 'licenses':
            return '''
                array(select name
                      from fossology_license
                      where id = ANY(
                         array_agg(%s.license_id))) as licenses''' % main_table
        return key

    def _get_from_list(self, table, ids, cols, cur=None, id_col='id'):
        """Fetches entries from the `table` such that their `id` field
        (or whatever is given to `id_col`) is in `ids`.
        Returns the columns `cols`.
        The `cur` parameter is used to connect to the database.
        """
        cur = self._cursor(cur)
        keys = map(self._convert_key, cols)
        query = """
            select {keys}
            from (values %s) as t(id)
            inner join {table} c
                on c.{id_col}=t.id
            inner join indexer_configuration i
                on c.indexer_configuration_id=i.id;
            """.format(
                keys=', '.join(keys),
                id_col=id_col,
                table=table)
        yield from execute_values_generator(
            cur, query,
            ((_id,) for _id in ids)
        )

    content_indexer_names = {
        'mimetype': 'content_mimetype',
        'fossology_license': 'content_fossology_license',
    }

    def content_get_range(self, content_type, start, end,
                          indexer_configuration_id, limit=1000,
                          with_textual_data=False, cur=None):
        """Retrieve contents with content_type, within range [start, end]
           bound by limit and associated to the given indexer
           configuration id.

           When asking to work on textual content, that filters on the
           mimetype table with any mimetype that is not binary.

        """
        cur = self._cursor(cur)
        table = self.content_indexer_names[content_type]
        if with_textual_data:
            extra = """inner join content_mimetype cm
                         on (t.id=cm.id and cm.mimetype like 'text/%%')"""
        else:
            extra = ""
        query = """select t.id
                   from %s t
                   inner join indexer_configuration ic
                   on t.indexer_configuration_id=ic.id
                   %s
                   where ic.id=%%s and
                         %%s <= t.id and t.id <= %%s
                   order by t.indexer_configuration_id, t.id
                   limit %%s""" % (table, extra)
        cur.execute(query, (indexer_configuration_id, start, end, limit))
        yield from cur

    def content_mimetype_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            'content_mimetype', ids, self.content_mimetype_cols, cur=cur)

    content_language_hash_keys = ['id', 'indexer_configuration_id']

    def content_language_missing_from_list(self, languages, cur=None):
        """List missing languages.

        """
        yield from self._missing_from_list(
            'content_language', languages, self.content_language_hash_keys,
            cur=cur)

    content_language_cols = [
        'id', 'lang',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    @stored_procedure('swh_mktemp_content_language')
    def mktemp_content_language(self, cur=None): pass

    def content_language_add_from_temp(self, conflict_update, cur=None):
        self._cursor(cur).execute("SELECT swh_content_language_add(%s)",
                                  (conflict_update, ))

    def content_language_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            'content_language', ids, self.content_language_cols, cur=cur)

    content_ctags_hash_keys = ['id', 'indexer_configuration_id']

    def content_ctags_missing_from_list(self, ctags, cur=None):
        """List missing ctags.

        """
        yield from self._missing_from_list(
            'content_ctags', ctags, self.content_ctags_hash_keys,
            cur=cur)

    content_ctags_cols = [
        'id', 'name', 'kind', 'line', 'lang',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    @stored_procedure('swh_mktemp_content_ctags')
    def mktemp_content_ctags(self, cur=None): pass

    def content_ctags_add_from_temp(self, conflict_update, cur=None):
        self._cursor(cur).execute("SELECT swh_content_ctags_add(%s)",
                                  (conflict_update, ))

    def content_ctags_get_from_list(self, ids, cur=None):
        cur = self._cursor(cur)
        keys = map(self._convert_key, self.content_ctags_cols)
        yield from execute_values_generator(
            cur, """
            select %s
            from (values %%s) as t(id)
            inner join content_ctags c
                on c.id=t.id
            inner join indexer_configuration i
                on c.indexer_configuration_id=i.id
            order by line
            """ % ', '.join(keys),
            ((_id,) for _id in ids)
         )

    def content_ctags_search(self, expression, last_sha1, limit, cur=None):
        cur = self._cursor(cur)
        if not last_sha1:
            query = """SELECT %s
                       FROM swh_content_ctags_search(%%s, %%s)""" % (
                           ','.join(self.content_ctags_cols))
            cur.execute(query, (expression, limit))
        else:
            if last_sha1 and isinstance(last_sha1, bytes):
                last_sha1 = '\\x%s' % hashutil.hash_to_hex(last_sha1)
            elif last_sha1:
                last_sha1 = '\\x%s' % last_sha1

            query = """SELECT %s
                       FROM swh_content_ctags_search(%%s, %%s, %%s)""" % (
                           ','.join(self.content_ctags_cols))
            cur.execute(query, (expression, limit, last_sha1))

        yield from cur

    content_fossology_license_cols = [
        'id', 'tool_id', 'tool_name', 'tool_version', 'tool_configuration',
        'licenses']

    @stored_procedure('swh_mktemp_content_fossology_license')
    def mktemp_content_fossology_license(self, cur=None): pass

    def content_fossology_license_add_from_temp(self, conflict_update,
                                                cur=None):
        """Add new licenses per content.

        """
        self._cursor(cur).execute(
            "SELECT swh_content_fossology_license_add(%s)",
            (conflict_update, ))

    def content_fossology_license_get_from_list(self, ids, cur=None):
        """Retrieve licenses per id.

        """
        cur = self._cursor(cur)
        keys = map(self._convert_key, self.content_fossology_license_cols)
        yield from execute_values_generator(
            cur, """
            select %s
            from (values %%s) as t(id)
            inner join content_fossology_license c on t.id=c.id
            inner join indexer_configuration i
                on i.id=c.indexer_configuration_id
            group by c.id, i.id, i.tool_name, i.tool_version,
                     i.tool_configuration;
            """ % ', '.join(keys),
            ((_id,) for _id in ids)
        )

    content_metadata_hash_keys = ['id', 'indexer_configuration_id']

    def content_metadata_missing_from_list(self, metadata, cur=None):
        """List missing metadata.

        """
        yield from self._missing_from_list(
            'content_metadata', metadata, self.content_metadata_hash_keys,
            cur=cur)

    content_metadata_cols = [
        'id', 'metadata',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    @stored_procedure('swh_mktemp_content_metadata')
    def mktemp_content_metadata(self, cur=None): pass

    def content_metadata_add_from_temp(self, conflict_update, cur=None):
        self._cursor(cur).execute("SELECT swh_content_metadata_add(%s)",
                                  (conflict_update, ))

    def content_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            'content_metadata', ids, self.content_metadata_cols, cur=cur)

    revision_intrinsic_metadata_hash_keys = [
        'id', 'indexer_configuration_id']

    def revision_intrinsic_metadata_missing_from_list(
            self, metadata, cur=None):
        """List missing metadata.

        """
        yield from self._missing_from_list(
            'revision_intrinsic_metadata', metadata,
            self.revision_intrinsic_metadata_hash_keys, cur=cur)

    revision_intrinsic_metadata_cols = [
        'id', 'metadata', 'mappings',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    @stored_procedure('swh_mktemp_revision_intrinsic_metadata')
    def mktemp_revision_intrinsic_metadata(self, cur=None): pass

    def revision_intrinsic_metadata_add_from_temp(
            self, conflict_update, cur=None):
        self._cursor(cur).execute(
                "SELECT swh_revision_intrinsic_metadata_add(%s)",
                (conflict_update, ))

    def revision_intrinsic_metadata_delete(
            self, entries, cur=None):
        cur = self._cursor(cur)
        cur.execute(
                "DELETE from revision_intrinsic_metadata "
                "WHERE (id, indexer_configuration_id) IN "
                "   (VALUES %s)" % (', '.join('%s' for _ in entries)),
                tuple((e['id'], e['indexer_configuration_id'])
                      for e in entries),)

    def revision_intrinsic_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            'revision_intrinsic_metadata', ids,
            self.revision_intrinsic_metadata_cols, cur=cur)

    origin_intrinsic_metadata_cols = [
        'id', 'metadata', 'from_revision', 'mappings',
        'tool_id', 'tool_name', 'tool_version', 'tool_configuration']

    origin_intrinsic_metadata_regconfig = 'pg_catalog.simple'
    """The dictionary used to normalize 'metadata' and queries.
    'pg_catalog.simple' provides no stopword, so it should be suitable
    for proper names and non-English content.
    When updating this value, make sure to add a new index on
    origin_intrinsic_metadata.metadata."""

    @stored_procedure('swh_mktemp_origin_intrinsic_metadata')
    def mktemp_origin_intrinsic_metadata(self, cur=None): pass

    def origin_intrinsic_metadata_add_from_temp(
            self, conflict_update, cur=None):
        cur = self._cursor(cur)
        cur.execute(
                "SELECT swh_origin_intrinsic_metadata_add(%s)",
                (conflict_update, ))

    def origin_intrinsic_metadata_delete(
            self, entries, cur=None):
        cur = self._cursor(cur)
        cur.execute(
                "DELETE from origin_intrinsic_metadata "
                "WHERE (id, indexer_configuration_id) IN"
                "   (VALUES %s)" % (', '.join('%s' for _ in entries)),
                tuple((e['id'], e['indexer_configuration_id'])
                      for e in entries),)

    def origin_intrinsic_metadata_get_from_list(self, ids, cur=None):
        yield from self._get_from_list(
            'origin_intrinsic_metadata', ids,
            self.origin_intrinsic_metadata_cols, cur=cur,
            id_col='id')

    def origin_intrinsic_metadata_search_fulltext(self, terms, *, limit, cur):
        regconfig = self.origin_intrinsic_metadata_regconfig
        tsquery_template = ' && '.join("plainto_tsquery('%s', %%s)" % regconfig
                                       for _ in terms)
        tsquery_args = [(term,) for term in terms]
        keys = (self._convert_key(col, 'oim') for col in
                self.origin_intrinsic_metadata_cols)

        query = ("SELECT {keys} FROM origin_intrinsic_metadata AS oim "
                 "INNER JOIN indexer_configuration AS i "
                 "ON oim.indexer_configuration_id=i.id "
                 "JOIN LATERAL (SELECT {tsquery_template}) AS s(tsq) ON true "
                 "WHERE oim.metadata_tsvector @@ tsq "
                 "ORDER BY ts_rank(oim.metadata_tsvector, tsq, 1) DESC "
                 "LIMIT %s;"
                 ).format(keys=', '.join(keys),
                          tsquery_template=tsquery_template)
        cur.execute(query, tsquery_args + [limit])
        yield from cur

    def origin_intrinsic_metadata_search_by_producer(
            self, last, limit, ids_only, mappings, tool_ids, cur):
        if ids_only:
            keys = 'oim.id'
        else:
            keys = ', '.join((self._convert_key(col, 'oim') for col in
                              self.origin_intrinsic_metadata_cols))
        query_parts = [
            "SELECT %s" % keys,
            "FROM origin_intrinsic_metadata AS oim",
            "INNER JOIN indexer_configuration AS i",
            "ON oim.indexer_configuration_id=i.id",
        ]
        args = []

        where = []
        if last:
            where.append('oim.id > %s')
            args.append(last)
        if mappings is not None:
            where.append('oim.mappings && %s')
            args.append(mappings)
        if tool_ids is not None:
            where.append('oim.indexer_configuration_id = ANY(%s)')
            args.append(tool_ids)
        if where:
            query_parts.append('WHERE')
            query_parts.append(' AND '.join(where))

        if limit:
            query_parts.append('LIMIT %s')
            args.append(limit)

        cur.execute(' '.join(query_parts), args)
        yield from cur

    indexer_configuration_cols = ['id', 'tool_name', 'tool_version',
                                  'tool_configuration']

    @stored_procedure('swh_mktemp_indexer_configuration')
    def mktemp_indexer_configuration(self, cur=None):
        pass

    def indexer_configuration_add_from_temp(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("SELECT %s from swh_indexer_configuration_add()" % (
            ','.join(self.indexer_configuration_cols), ))
        yield from cur

    def indexer_configuration_get(self, tool_name,
                                  tool_version, tool_configuration, cur=None):
        cur = self._cursor(cur)
        cur.execute('''select %s
                       from indexer_configuration
                       where tool_name=%%s and
                             tool_version=%%s and
                             tool_configuration=%%s''' % (
                                 ','.join(self.indexer_configuration_cols)),
                    (tool_name, tool_version, tool_configuration))

        return cur.fetchone()



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import unittest

from typing import Any, Dict

from swh.indexer.mimetype import (
    MimetypeIndexer, MimetypeRangeIndexer, compute_mimetype_encoding
)

from swh.indexer.tests.utils import (
    CommonContentIndexerTest, CommonContentIndexerRangeTest,
    BASE_TEST_CONFIG, fill_storage, fill_obj_storage, filter_dict,
)


class BasicTest(unittest.TestCase):
    def test_compute_mimetype_encoding(self):
        """Compute mimetype encoding should return results"""
        for _input, _mimetype, _encoding in [
                ('du français'.encode(), 'text/plain', 'utf-8'),
                (b'def __init__(self):', 'text/x-python', 'us-ascii')]:

            actual_result = compute_mimetype_encoding(_input)
            self.assertEqual(actual_result, {
                'mimetype': _mimetype,
                'encoding': _encoding
            })


CONFIG = {
    **BASE_TEST_CONFIG,
    'tools': {
        'name': 'file',
        'version': '1:5.30-1+deb9u1',
        'configuration': {
            "type": "library",
            "debian-package": "python3-magic"
        },
    },
}  # type: Dict[str, Any]


class TestMimetypeIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Mimetype indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """
    legacy_get_format = True

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_mimetype_get(ids)

    def setUp(self):
        self.indexer = MimetypeIndexer(config=CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0 = '01c9379dfc33803963d07c1ccc748d3fe4c96bb5'
        self.id1 = '688a5ef812c53907562fe379d4b3851e69c7cb15'
        self.id2 = 'da39a3ee5e6b4b0d3255bfef95601890afd80709'

        tool = {k.replace('tool_', ''): v
                for (k, v) in self.indexer.tool.items()}

        self.expected_results = {
            self.id0: {
                'id': self.id0,
                'tool': tool,
                'mimetype': 'text/plain',
                'encoding': 'us-ascii',
            },
            self.id1: {
                'id': self.id1,
                'tool': tool,
                'mimetype': 'text/plain',
                'encoding': 'us-ascii',
            },
            self.id2: {
                'id': self.id2,
                'tool': tool,
                'mimetype': 'application/x-empty',
                'encoding': 'binary',
            }
        }


RANGE_CONFIG = dict(list(CONFIG.items()) + [('write_batch_size', 100)])


class TestMimetypeRangeIndexer(
        CommonContentIndexerRangeTest, unittest.TestCase):
    """Range Mimetype Indexer tests.

    - new data within range are indexed
    - no data outside a range are indexed
    - with filtering existing indexed data prior to compute new index
    - without filtering existing indexed data prior to compute new index

    """
    def setUp(self):
        super().setUp()
        self.indexer = MimetypeRangeIndexer(config=RANGE_CONFIG)
        self.indexer.catch_exceptions = False
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0 = '01c9379dfc33803963d07c1ccc748d3fe4c96bb5'
        self.id1 = '02fb2c89e14f7fab46701478c83779c7beb7b069'
        self.id2 = '103bc087db1d26afc3a0283f38663d081e9b01e6'
        tool_id = self.indexer.tool['id']

        self.expected_results = {
            self.id0: {
                'encoding': 'us-ascii',
                'id': self.id0,
                'indexer_configuration_id': tool_id,
                'mimetype': 'text/plain'},
            self.id1: {
                'encoding': 'us-ascii',
                'id': self.id1,
                'indexer_configuration_id': tool_id,
                'mimetype': 'text/x-python'},
            self.id2: {
                'encoding': 'us-ascii',
                'id': self.id2,
                'indexer_configuration_id': tool_id,
                'mimetype': 'text/plain'}
        }


def test_mimetype_w_no_tool():
    with pytest.raises(ValueError):
        MimetypeIndexer(config=filter_dict(CONFIG, 'tools'))


def test_mimetype_range_w_no_tool():
    with pytest.raises(ValueError):
        MimetypeRangeIndexer(config=filter_dict(CONFIG, 'tools'))



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI
from .base import JsonMapping


class NpmMapping(JsonMapping):
    """
    dedicated class for NPM (package.json) mapping and translation
    """
    name = 'npm'
    mapping = CROSSWALK_TABLE['NodeJS']
    filename = b'package.json'
    string_fields = ['name', 'version', 'homepage', 'description', 'email']

    _schema_shortcuts = {
            'github': 'git+https://github.com/%s.git',
            'gist': 'git+https://gist.github.com/%s.git',
            'gitlab': 'git+https://gitlab.com/%s.git',
            # Bitbucket supports both hg and git, and the shortcut does not
            # tell which one to use.
            # 'bitbucket': 'https://bitbucket.org/',
            }

    def normalize_repository(self, d):
        """https://docs.npmjs.com/files/package.json#repository

        >>> NpmMapping().normalize_repository({
        ...     'type': 'git',
        ...     'url': 'https://example.org/foo.git'
        ... })
        {'@id': 'git+https://example.org/foo.git'}
        >>> NpmMapping().normalize_repository(
        ...     'gitlab:foo/bar')
        {'@id': 'git+https://gitlab.com/foo/bar.git'}
        >>> NpmMapping().normalize_repository(
        ...     'foo/bar')
        {'@id': 'git+https://github.com/foo/bar.git'}
        """
        if isinstance(d, dict) and isinstance(d.get('type'), str) \
                and isinstance(d.get('url'), str):
            url = '{type}+{url}'.format(**d)
        elif isinstance(d, str):
            if '://' in d:
                url = d
            elif ':' in d:
                (schema, rest) = d.split(':', 1)
                if schema in self._schema_shortcuts:
                    url = self._schema_shortcuts[schema] % rest
                else:
                    return None
            else:
                url = self._schema_shortcuts['github'] % d

        else:
            return None

        return {'@id': url}

    def normalize_bugs(self, d):
        """https://docs.npmjs.com/files/package.json#bugs

        >>> NpmMapping().normalize_bugs({
        ...     'url': 'https://example.org/bugs/',
        ...     'email': 'bugs@example.org'
        ... })
        {'@id': 'https://example.org/bugs/'}
        >>> NpmMapping().normalize_bugs(
        ...     'https://example.org/bugs/')
        {'@id': 'https://example.org/bugs/'}
        """
        if isinstance(d, dict) and isinstance(d.get('url'), str):
            return {'@id': d['url']}
        elif isinstance(d, str):
            return {'@id': d}
        else:
            return None

    _parse_author = re.compile(r'^ *'
                               r'(?P<name>.*?)'
                               r'( +<(?P<email>.*)>)?'
                               r'( +\((?P<url>.*)\))?'
                               r' *$')

    def normalize_author(self, d):
        """https://docs.npmjs.com/files/package.json#people-fields-author-contributors'

        >>> from pprint import pprint
        >>> pprint(NpmMapping().normalize_author({
        ...     'name': 'John Doe',
        ...     'email': 'john.doe@example.org',
        ...     'url': 'https://example.org/~john.doe',
        ... }))
        {'@list': [{'@type': 'http://schema.org/Person',
                    'http://schema.org/email': 'john.doe@example.org',
                    'http://schema.org/name': 'John Doe',
                    'http://schema.org/url': {'@id': 'https://example.org/~john.doe'}}]}
        >>> pprint(NpmMapping().normalize_author(
        ...     'John Doe <john.doe@example.org> (https://example.org/~john.doe)'
        ... ))
        {'@list': [{'@type': 'http://schema.org/Person',
                    'http://schema.org/email': 'john.doe@example.org',
                    'http://schema.org/name': 'John Doe',
                    'http://schema.org/url': {'@id': 'https://example.org/~john.doe'}}]}
        """ # noqa
        author = {'@type': SCHEMA_URI+'Person'}
        if isinstance(d, dict):
            name = d.get('name', None)
            email = d.get('email', None)
            url = d.get('url', None)
        elif isinstance(d, str):
            match = self._parse_author.match(d)
            if not match:
                return None
            name = match.group('name')
            email = match.group('email')
            url = match.group('url')
        else:
            return None
        if name and isinstance(name, str):
            author[SCHEMA_URI+'name'] = name
        if email and isinstance(email, str):
            author[SCHEMA_URI+'email'] = email
        if url and isinstance(url, str):
            author[SCHEMA_URI+'url'] = {'@id': url}
        return {"@list": [author]}

    def normalize_license(self, s):
        """https://docs.npmjs.com/files/package.json#license

        >>> NpmMapping().normalize_license('MIT')
        {'@id': 'https://spdx.org/licenses/MIT'}
        """
        if isinstance(s, str):
            return {"@id": "https://spdx.org/licenses/" + s}

    def normalize_homepage(self, s):
        """https://docs.npmjs.com/files/package.json#homepage

        >>> NpmMapping().normalize_homepage('https://example.org/~john.doe')
        {'@id': 'https://example.org/~john.doe'}
        """
        if isinstance(s, str):
            return {"@id": s}

    def normalize_keywords(self, l):
        """https://docs.npmjs.com/files/package.json#homepage

        >>> NpmMapping().normalize_keywords(['foo', 'bar'])
        ['foo', 'bar']
        """
        if isinstance(l, list):
            return [x for x in l if isinstance(x, str)]
